/* @(#) SupportPacks/ms0c/public/mqsecche.h, supportpacks, p000, p000-L000330          1.3  99/05/22 22:34:20 */
#ifndef _MQSECCHE_H
#define _MQSECCHE_H
/*---------------------------------------------------------------
 *  Author(s)     : Benjamin Y. Goldshlager
 *
 *  Description:
 *      This the header file for the MQ Channel Security Exit.
 *
 *      This code is base on amqdcexc.c Sample
 *
 *
 *  Notes:
 *
 *  Change Log:
 *    $Log$
 *
 *    Who  Date      Defect  Description
 *    ---  --------  ------  ------------------------------------
 *    BYG  01/26/99          Add DEBUG_MAX_APP_DATA option
 *
 *--------------------------------------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>

#include <cmqc.h>
#include <cmqxc.h>

//#define dtkError      printf
//#define dtkDebug      printf
//#define dtkWarning    printf
//#define dtkDisplay    printf

#if MQAT_DEFAULT==MQAT_UNIX
  #include <errno.h>
  #include <sys/types.h>
  #include <unistd.h>
  #include <netinet/in.h>
#endif

#if MQAT_DEFAULT==MQAT_WINDOWS_NT
  //#define  TRUE        1
  /* htons is present in later versions of the Microsoft Development Environment eg v5.0.*/
  /* When these are the only ones supported                                              */
  /* the next line and all subsequent references to htons1 can be removed. As currently  */
  /* the code must compile on both older and newer versions of the development           */
  /* environment we need to rename htons to htons1.                                      */
  #define  htons1(x)    (((x<<8)&0xff00) | ((x>>8)&0xff))
  #include <windows.h>
#endif

#if MQAT_DEFAULT==MQAT_OS2
  //#define  TRUE        1
  #define  htons(x)    (((x<<8)&0xff00) | ((x>>8)&0xff))
  /*BYG*/
  #ifdef GSS_ENTRUST_API
    #error Entrust GSS-API not supported on OS2
  #endif
    #include <dce/pthread.h>       /* DCE threads                         */
#endif

/*BYG*/
#ifdef GSS_DCE_API
  #include <dce/gssapi.h>
#else
  #include <gssapi.h>
  #include <etsesn.h>
#endif

#ifndef BOOL
#define BOOL    char
#endif

#ifndef TRUE
#define TRUE    1
#define FALSE   0
#endif

#if MQAT_DEFAULT==MQAT_UNIX
#define WORD    int
#endif

typedef unsigned short OM_uint16;
	       /* Local Structure definitions */
typedef struct
{
   MQBYTE        expectedSecMsg;
   //PMQBYTE       pTargetName;
   MQCHAR48      mqTargetName;
   gss_name_t    PartnerGSSName;
   gss_cred_id_t credentialHandle;
} SECEXITDATA, MQPOINTER PSECEXITDATA;

/*********************************************************************/
/*                                                                   */
/*    WARNING    WARNING   WARNING   WARNING   WARNING   WARNING     */
/*                                                                   */
/* In the struct below MQ development have had to add 5 bytes of     */
/* padding.                                                          */
/* This is there to overcome the fact that the components of MQSeries*/
/* for NT are compiled with a switch that aligns structs on a 8 byte */
/* boundary. Adding these extra bytes means that the alignment is    */
/* kept correct for other platforms which do not align on an 8 byte  */
/* boundary.                                                         */
/* If a compile option with different alignment is used then this    */
/* struct may well need changing to suit.                            */
/*                                                                   */
/*********************************************************************/
typedef struct
{
   struct
   {
      OM_uint16      mechanism;
      MQBYTE         msgType;
      MQBYTE         padding;
      MQBYTE         padding1;
      MQBYTE         padding2;
      MQBYTE         padding3;
      MQBYTE         padding4;
   }        header;
   MQBYTE   msg;
} GSSSECMSG, MQPOINTER PGSSSECMSG;



typedef struct
{
   OM_uint16      mechanism;
   MQBYTE         sender;
   gss_ctx_id_t   contextHandle;
} CONTEXTDATA, MQPOINTER PCONTEXTDATA;

/* Exit instance local variables - These variables are set from */
/* the INI file.                                                */

/* Define some defaults */

// Security Mechanism
#define SPKM_1_STRING           "gss_mech_spkm1"
#define SPKM_2_STRING           "gss_mech_spkm2"
#define SPKM_DEFAULT_STRING     "GSS_C_NULL_OID_SET"

// Symetric Key exchange protocol
#define PROTOCOL_RSA_STRING             "SPKM_KEY_ESTB_PROTOCOL_RSA"
#define PROTOCOL_D_H_STRING             "SPKM_KEY_ESTB_PROTOCOL_D_H"
#define PROTOCOL_DEFAULT_STRING "SPKM_KEY_ESTB_PROTOCOL_DEFAULT"
//#define DEFAULT_SYM_KEY_EXCHANGE_PROTOCOL "SPKM_KEY_ESTB_PROTOCOL_DEFAULT"

// File names - edit these if you wish, but they can all be overridden. The
// Channel Ini file is named in the ExitData component of the channel
// definition, and that in turn can specify all the other files.
#if MQAT_DEFAULT==MQAT_WINDOWS_NT
  #define DEFAULT_LOG_FILE                "c:\\temp\\mqsecche.log"
  #define DEFAULT_CHANNEL_INI_FILE        "c:\\winnt\\profiles\\mqsecche.ini"
  #define DEFAULT_KEYTAB_FILE             "c:\\winnt\\profiles\\mqm\\keytab"
  #define DEFAULT_ENTRUST_INI_FILE        "c:\\winnt\\entrust.ini"
  #define DEFAULT_ENTRUST_EPF_FILE        "c:\\winnt\\profiles\\mqm\\mqm.epf"
#else
  #define DEFAULT_LOG_FILE                "/tmp/mqsecche.log"
  #define DEFAULT_CHANNEL_INI_FILE        "/u/mqm/.mqsecche.ini"
  #define DEFAULT_KEYTAB_FILE             "/u/mqm/.keytab"
  #define DEFAULT_ENTRUST_INI_FILE        "/u/mqm/entrust.ini"
  #define DEFAULT_ENTRUST_EPF_FILE        "/u/mqm/.mqm.epf"
#endif

#ifndef FILENAME_MAX
#define FILENAME_MAX    260
#endif

#define MAX_MSG_LEN             1024

/* Debug values/flags */
#define EAMS_DEBUG_NONE_STRING          "NONE"
#define EAMS_DEBUG_FILE_STRING          "FILE"
#define EAMS_DEBUG_CONSOLE_STRING       "CONSOLE"

#define EAMS_DEBUG_NONE    0x00
#define EAMS_DEBUG_FILE    0x01
#define EAMS_DEBUG_CONSOLE 0x02

typedef struct
{
  // Authorized partner's Queue Manager name
  MQCHAR          mqTargetName[MQ_Q_MGR_NAME_LENGTH+1];

  // Authorized partner's Distinguished Name
  char*           pszTargetNameDN;

  // Local Queue Manager Name
  MQCHAR          szLocalQMgrName[MQ_Q_MGR_NAME_LENGTH+1];
  // Log file name
  //MQCHAR                szLogFile[FILENAME_MAX+1];

  // Keytab file name
  MQCHAR          szKeyTab[FILENAME_MAX+1];

  // Entrust INI file name
  MQCHAR          szEntrustINIFile[FILENAME_MAX+1];

  // Entrust User Profile file name
  MQCHAR          szEntrustEPFFile[FILENAME_MAX+1];

  // Channel Exit INI file name.  32 are the limit enforced by MQ
  MQCHAR          szChannelExitINIFile[MQ_EXIT_DATA_LENGTH+1];

  // Debug file name - defaults to szLogFile.
  MQCHAR          szDebugFile[FILENAME_MAX+1];

  // Debug options. EAMS_DEBUG_NONE, EAMS_DEBUG_FILE, EAMS_DEBUG_CONSOLE
  MQBYTE          cDebug;

  // Number of application message to display when debugging
  // DEBUG_MAX_APP_DATA option
  int                     iDebugMaxAppData;

  // Security machanism.  Valid values 'gss_mech_spkm1' or 'gss_mech_spkm2'
  gss_OID         entSecurityMechanism;

  // Symetric key exchange protocol
  ents_key_estb_protocol entSymmetricKeyExchangeProtocol;

  // Context life time - 3600 seconds is the default
  OM_uint32       ulReqContextLifeTime;

  // Encryption option.- default 0 (NO).
  MQBYTE          bEncrypt;

  // Number of times initSecContext has been called
  int                     iInitSecContextCounter;

  // Number of times acceptSecContext has been called
  int                     iAcceptSecContextCounter;

  // Number of times Message Exit was called to wrap data
  unsigned long           ulWrapMessagesCounter;

  // Number of times Message Exit was called to unwrap data
  unsigned long           ulUnwrapMessagesCounter;

} EXITLOCALVARS, MQPOINTER PEXITLOCALVARS;

/*
 * The EXITUSERAREA structure is passed in the MQCXP Structure
 * (Channel Exit Parameter)'s pChParm->ExitUserArea.
 *
 * NB. The ExitUserArea is defined to be 16 bytes long. Enough to contain
 * two pointers even if it were on a 64-bit system. First time through, this
 * will be empty (all bytes = 0) so the pointers will be NULL.
 */
typedef struct
{
  PSECEXITDATA   pSecExitData;    // Security stuff
  PEXITLOCALVARS pExitLocalVars;  // Local variables of this channel exit instance
} EXITUSERAREA, MQPOINTER PEXITUSERAREA;


/* Local macros                */

#define SECURITYEXITCALL   pParms->ExitId==MQXT_CHANNEL_SEC_EXIT
#define SENDEXITCALL       pParms->ExitId==MQXT_CHANNEL_SEND_EXIT
#define RECEIVEEXITCALL    pParms->ExitId==MQXT_CHANNEL_RCV_EXIT
#define MSGEXITCALL        pParms->ExitId==MQXT_CHANNEL_MSG_EXIT

#define EXPECTINGMSG(x)    ((x==MQXCC_SEND_AND_REQUEST_SEC_MSG)||(x==MQXCC_OK))
#define EXPECTSECMSG(x,y,z) y->expectedSecMsg=EXPECTINGMSG(x)?z:GSS_SEC_NONE

#define UNEXPECTEDMSG(x,y) (x->expectedSecMsg!=y)
#define IAMIDLE(x)         (x->expectedSecMsg==GSS_SEC_NONE)

#define GSS_DCE_MACHANISM     (OM_uint16) 0x1d76
/*BYG*/
#define GSS_ENTRUST_MECHANISM (OM_uint16) 0xa1b2

#if MQAT_DEFAULT==MQAT_WINDOWS_NT
  #if GSS_DCE_API
    #define SEC_MECHANISM(x)  (x==htons1(GSS_DCE_MACHANISM))
  #else   // GSS_ENTRUST_API
    #define SEC_MECHANISM(x)  (x==htons1(GSS_ENTRUST_MECHANISM))
  #endif
#else
  #if GSS_DCE_API
    #define SEC_MECHANISM(x)  (x==htons(GSS_DCE_MACHANISM))
  #else   // GSS_ENTRUST_API
    #define SEC_MECHANISM(x)  (x==htons(GSS_ENTRUST_MECHANISM))
  #endif
#endif

/* Types of message flowed to the partner exit */

#define GSS_SEC_NONE         0
#define GSS_SEC_INIT_SEC     1
#define GSS_SEC_ACC_SEC      2
#define GSS_SEC_SEAL         3

/* Utlities prototypes */
#include "mqsecutl.h"

/* GSS Routine prototypes */
#include "mqsecrtn.h"

#endif          // _MQSECCHE_H
