static char *sccsid = "@(#) SupportPacks/ms0c/public/mqsecrtn.c, supportpacks, p000, p000-L000330          1.5  00/03/06 16:55:06";
/*---------------------------------------------------------------
 *  Author(s)     : Benjamin Y. Goldshlager
 *
 *  Name: MQSECRTN.C
 *
 *  Description:
 *      MQ Security Channel Exit routines.
 *
 *      This code is base on amqdcexc.c sample
 *
 *
 *  Notes:
 *
 *  Change Log:
 *    $Log$
 *
 *    Who  Date      Defect  Description
 *    ---  --------  ------  ------------------------------------
 *    BYG  01/13/99          Add debug in terminateSecurityExit()
 *    BYG  01/15/99          In acceptSecContext() move compareDNNames()
 *                             before composeSecTokenMsg()
 *    BYG  01/26/99          Add DEBUG_MAX_APP_DATA option
 *    BYG  04/13/99          Fix logic to display correctly msg data when
 *                             using Message Exit.
 *--------------------------------------------------------------*/

#include "mqsecche.h"

/* IsStatusOK() constants */
#define CE_ERROR_MSG_TYPE                1
#define CE_WARNING_MSG_TYPE              2
#define CE_DEBUG_MSG_TYPE                3
//#define SRM_MAX_DEBUG_STRING  20

static char* perror2GSS (PEXITLOCALVARS pExitLocalVars,
	     int iStatusType,
	     OM_uint32 gssStatus);

static void printBinaryBuffer (PEXITLOCALVARS pExitLocalVars,
	         char *pre_s,
	         int wrap_width,
	         gss_buffer_desc buf);

static void printGSSBufferDesc (PEXITLOCALVARS pExitLocalVars,
	        char *pre_s,
	        char * label,
	        gss_buffer_desc buf);

static void displayCertificate (PEXITLOCALVARS pExitLocalVars,
	        char *pre_s,
	        ents_cert_id_t cert,
	        int enc);

static void displayCredCertificates (PEXITLOCALVARS pExitLocalVars,
	           char *pre_s,
	           gss_cred_id_t cred);

static void displayPeerCertificates (PEXITLOCALVARS pExitLocalVars,
	           char *pre_s,
	           gss_ctx_id_t ctxt);

static char *oid_to_mech_str(gss_OID mech);


/*
 * Display the status message in an error from the GSS-API
 */

static char* perror2GSS (PEXITLOCALVARS pExitLocalVars,
	     int iStatusType,               // Type of status requested
	       OM_uint32 gssStatus)       // Status requrested)
{
  OM_uint32       gssMajor, gssMinor;
  OM_uint32       gssMsgCtxt=0;
  gss_buffer_desc gssBuf;

  char           *pszMsg = 0, *pszBuff;
  int             iMsgLen = 0;

  do
  {
    gssMajor = gss_display_status (&gssMinor,
	    gssStatus,      /* Status to display */
	    iStatusType,    /* Status type */
	    GSS_C_NULL_OID, /* Default mechanism */
	    &gssMsgCtxt,    /* Message context */
	    &gssBuf);       /* Output error message */

    if (gssMajor != GSS_S_COMPLETE)
    {
      dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> perror2GSS(): "
	    "Cannot display GSS-%s error %d\n", __FILE__, __LINE__,
	   iStatusType == GSS_C_GSS_CODE? "API" : "API mechanism", gssStatus);

      if (iMsgLen)
      {
	free (pszMsg);
	pszMsg = NULL;
      }

      break;
    }

    //dtkError ("%s: %.*s\n", pszMsg, gssBuf.length, gssBuf.value);

    if (iMsgLen == 0)
    {
      iMsgLen = gssBuf.length + 1;
      pszMsg = calloc (iMsgLen, 1);
      memcpy (pszMsg, gssBuf.value, gssBuf.length);

      dtkDebug (pExitLocalVars, "perror2GSS(): Allocated message buffer for "
	"<%ld> bytes\n", iMsgLen);
    }
    else
    {
      dtkDebug (pExitLocalVars, "perror2GSS(): Freeing pszMsg "
	    "of size <%ld>\n", iMsgLen);
      iMsgLen += gssBuf.length + 1;

      pszBuff = calloc (iMsgLen, 1);
      strcpy (pszBuff, pszMsg);
      free (pszMsg);

      memcpy (pszBuff + strlen (pszBuff), gssBuf.value, gssBuf.length);
      pszMsg = pszBuff;
      dtkDebug (pExitLocalVars, "perror2GSS(): Re-Allocated message buffer for "
	"<%ld> bytes\n", iMsgLen);
    }

    gss_release_buffer (&gssMinor, &gssBuf);

  } while (gssMsgCtxt);

  return pszMsg;
}


/*
 * Display error status message from GSS-API
 */
/**********************************************************************/
/*                                                                    */
/* Function Name: convertGSSNameToString                              */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*       Called to display error status message from GSS-API          */
/*                                                                    */
/* Input Parameters:  pExitLocalVars    - Local variables             */
/*                    uiMajor           - GSS major status            */
/*                    uiMinor           - GSS minor status            */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns:           char*  - On success contains error string of    */
/*                             GSS error.  The string is allocated    */
/*                             from the heap, which must be freed by  */
/*                             caller after use.                      */
/*                                                                    */
/**********************************************************************/

char* perrorGSS (PEXITLOCALVARS pExitLocalVars,
	 OM_uint32 uiMajor,
	 OM_uint32 uiMinor)
{
  char *pszFullMsg = NULL, *pszMajorMsg = NULL, *pszMinorMsg = NULL;

  /* If the major status value is GSS_S_FAILURE, just print the minor
     status value -- unless there is no minor status value, in which
       case do print the major status value */

  if (uiMajor != GSS_S_FAILURE || uiMinor == GSS_S_G_NONE)
    pszMajorMsg = perror2GSS (pExitLocalVars, GSS_C_GSS_CODE, uiMajor);

  /* If there is a minor status value, print it */

  if (uiMinor != GSS_S_G_NONE)
  {
    pszMinorMsg = perror2GSS (pExitLocalVars, GSS_C_MECH_CODE, uiMinor);
    if (pszMinorMsg)
    {       // Handle minor status error text
      if (pszMajorMsg)
      { // If there was major status error text, get minor status
	// and concatenate it to major status.  Extra 1 byte for blank separator
	int iLen = strlen (pszMajorMsg) + strlen (pszMinorMsg) + 2;

	dtkDebug (pExitLocalVars, "perrorGSS(): Allocating buffer size "
	  "<%ld> for both minor and major status messages\n", iLen + 1);

	pszFullMsg = calloc (iLen, 1);
	if (pszMajorMsg)
	{
	  strcpy (pszFullMsg, pszMajorMsg);
	  strcat (pszFullMsg, " ");

	  dtkDebug (pExitLocalVars, "perrorGSS(): Freeing pszMajorMsg of size "
	    "<%ld> \n", strlen (pszMajorMsg) + 1);
	  free (pszMajorMsg);
	}

	strcat (pszFullMsg, pszMinorMsg);

	dtkDebug (pExitLocalVars, "perrorGSS(): Freeing pszMinorMsg of size "
	  "<%ld> \n", strlen (pszMinorMsg) + 1);
	free (pszMinorMsg);
      }
      else
      {       // Only minor state error text is returned.
	pszFullMsg = pszMinorMsg;
      }
    }
    // Could not get minor status error text
  }
  else
  {       // No minor status error text retrieved
    pszFullMsg = pszMajorMsg;
  }

  return pszFullMsg;
}


/**********************************************************************/
/*                                                                    */
/* Function Name: convertGSSNameToString                              */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*       Called to convert GSS name to null delimited string          */
/*                                                                    */
/* Input Parameters:  pExitLocalVars    - Local variables             */
/*                    internalGSSName   - GSS name                    */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns:           char*  - On success contains string of GSS name */
/*                             This string is allocated from the heap */
/*                             must be freed by caller after use.     */
/*                                                                    */
/*                             A NULL is returned on failure          */
/*                                                                    */
/**********************************************************************/

char *convertGSSNameToString (PEXITLOCALVARS pExitLocalVars,
	        gss_name_t internalGSSName)
{
  OM_uint32               ulMajor  = 0;
  OM_uint32               ulMinor  = 0;
  gss_OID                 nameType = GSS_C_NULL_OID;
  gss_buffer_desc fullGSSName = {0, NULL};
  char*                   pszDNName = NULL;

  if (internalGSSName == NULL)
    return NULL;

  ulMajor = gss_display_name (&ulMinor, internalGSSName,
	        &fullGSSName, &nameType);

  if (!IsStatusOK (pExitLocalVars, __FILE__, __LINE__, ulMajor, ulMinor,
    GSS_S_COMPLETE, GSS_S_G_NONE, "convertGSSNameToString()->gss_display_name()", TRUE))
    return NULL;

  dtkDebug (pExitLocalVars, "convertGSSNameToString(): Allocated message buffer for "
	"<%ld> bytes\n", fullGSSName.length+1);

  pszDNName = calloc (fullGSSName.length+1, 1);
  if (pszDNName == NULL)
  {
    dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> convertGSSNameToString(): "
	  "Could not allocate memory\n", __FILE__, __LINE__);
    return NULL;
  }

  memcpy (pszDNName, fullGSSName.value, fullGSSName.length);
  gss_release_buffer (&ulMinor, &fullGSSName);

  return pszDNName;
}


/**********************************************************************/
/*                                                                    */
/* Function Name: IsStatusOK                                          */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*       Called to check for GSS-API status and display error or      */
/*       warning messages.  This routine displayes GSS error if       */
/*       pszMsgPrefix  != NULL.  In either case TRUE is returned      */
/*       for success, and FALSE for failure.                          */
/*                                                                    */
/* Input Parameters:  pExitLocalVars    - Local variables             */
/*                    pszSource         - Source filename             */
/*                    gssMajorStatus    - GSS major status            */
/*                    gssMinorStatus    - GSS minor status            */
/*                    gssExpectedMajorStatus - Expected major status  */
/*                    gssExpectedMinorStatus - Expected minor status  */
/*                    pszMsgPrefix          - Message prefix          */
/*                    bWarn             - Used in a case where API    */
/*                                        returns a warning.          */
/*                                 TRUE  - display as warning message */
/*                                 FALSE - display as debug message   */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns:                       TRUE  - status matched expectation              */
/*                    FALSE - status did not match expectation        */
/*                                                                    */
/**********************************************************************/

BOOL IsStatusOK (PEXITLOCALVARS pExitLocalVars,
	 char           *pszSource,
	 int            iLine,
	 OM_uint32      gssMajorStatus,
	 OM_uint32      gssMinorStatus,
	 OM_uint32      gssExpectedMajorStatus,
	 OM_uint32      gssExpectedMinorStatus,
	 char           *pszMsgPrefix,
	 BOOL           bWarn)
{
  char szMsg[MAX_MSG_LEN+1];
  char *pszGSSErrorMsg;
  int  iMsgType = 0;
  BOOL rc;

  if (gssMajorStatus == GSS_S_COMPLETE)
  {
    rc = TRUE;
  }
  else if ((gssMajorStatus != gssExpectedMajorStatus  &&
	gssMinorStatus != gssExpectedMinorStatus) ||
       (gssMajorStatus == gssExpectedMajorStatus  &&
	gssMinorStatus != gssExpectedMinorStatus) ||
       (gssMajorStatus != gssExpectedMajorStatus  &&
	gssMinorStatus == gssExpectedMinorStatus))

  { // Error if gssMajorStatus and gssMinorStatus are not acceptable
    // or gssMajorStatus is acceptable but gssMinorStatus is not
    rc = FALSE;
    iMsgType = CE_ERROR_MSG_TYPE;
  }
  else
  { // Even though major and minor status are acceptable, issue a
    // warning if anything is not perfect.
    rc = TRUE;
    if (bWarn)
      iMsgType = CE_WARNING_MSG_TYPE;
    else
      iMsgType = CE_DEBUG_MSG_TYPE;
  }

  if (iMsgType > 0 && pszMsgPrefix)
  { /* Major and Minor status code are not acceptable (caller's GSS call failed) */
    pszGSSErrorMsg = perrorGSS (pExitLocalVars, gssMajorStatus, gssMinorStatus);

    if (pszSource)  // If available, display source and line number
      sprintf (szMsg, "SourceFile <%s> SourceLine <%d> "
	"%s GSS-API reported: <%s>\n",  pszSource, iLine,
	pszMsgPrefix ? pszMsgPrefix : "Error from unknown routine: ",
	pszGSSErrorMsg ? pszGSSErrorMsg : "could not retrieve error");
    else
      sprintf (szMsg, "%s GSS-API reported: <%s>\n",
	pszMsgPrefix ? pszMsgPrefix : "Error from unknown routine: ",
	pszGSSErrorMsg ? pszGSSErrorMsg : "could not retrieve error");

    switch (iMsgType)
    {
      case CE_ERROR_MSG_TYPE:
	dtkError (pExitLocalVars, szMsg);
	break;
      case CE_WARNING_MSG_TYPE:
	dtkWarning (pExitLocalVars, szMsg);
	break;
      default: // CE_DEBUG_MSG_TYPE
	dtkDebug (pExitLocalVars, szMsg);
    }

    if (pszGSSErrorMsg)
    {
      dtkDebug (pExitLocalVars, "IsStatusOK(): Freeing pszGSSErrorMsg of size <%ld>\n",
	  strlen (pszGSSErrorMsg) + 1);

      free (pszGSSErrorMsg);
    }
  }

  return rc;
}


/*
 * Display a binary buffer as hexadecimal with wrap-around
 */

static void printBinaryBuffer (PEXITLOCALVARS pExitLocalVars,
	         char *pre_s,
	         int wrap_width,
	         gss_buffer_desc buf)
{
  unsigned int i = 0, buf_length = buf.length;
  unsigned char * ch = (unsigned char*)buf.value ;

  wrap_width /= 2;  /* two printed characters per binary byte */
  if (buf_length == 0)
    return ;

  while (i < buf_length)
  {
    if (!(i % wrap_width))
      dtkDebug2 (pExitLocalVars, "%s: -- ", pre_s);
    dtkDebug2 (pExitLocalVars, "%2.2X", (unsigned int) ch[i++]);
    if (!(i % wrap_width))
      dtkDebug2 (pExitLocalVars, "\n");
  }
  if (i % wrap_width)
    dtkDebug2 (pExitLocalVars, "\n");
}


/*
 * Display the character string contained in a gss_buffer_desc
 */

static void printGSSBufferDesc (PEXITLOCALVARS pExitLocalVars,
	        char *pre_s,
	        char * label,
	        gss_buffer_desc buf)
{
   if (buf.length != 0)
      dtkDebug2 (pExitLocalVars, "%s: %s %.*s\n",
	pre_s, label, buf.length, (char *) buf.value);
}


/**********************************************************************/
/*                                                                    */
/* Function Name: displayCertificate                                  */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*       Called to display the information in the certificate         */
/*                                                                    */
/* Input Parameters:  pExitLocalVars    - Local variables             */
/*                    pre_s             - Display message prefix      */
/*                    cert              - Security credentials        */
/*                    enc               - Display common fields flag  */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns:           None                                            */
/*                                                                    */
/**********************************************************************/

static void displayCertificate (PEXITLOCALVARS pExitLocalVars,
	        char *pre_s,
	        ents_cert_id_t cert,
	        int enc)
{
  OM_uint32 status, minor;
  gss_buffer_desc ca_name_buf, owner_name_buf ;
  ents_cert_info_t cert_info ;

  status = ents_get_certificate_info (&minor, cert, &cert_info);
  if (!IsStatusOK (pExitLocalVars, __FILE__, __LINE__, status, minor,
    GSS_S_COMPLETE, GSS_S_G_NONE,
    "displayCertificate()->ents_get_certificate_info()", TRUE))
  {
    return;
  }

  gss_display_name (&minor, cert_info->ca_name,    &ca_name_buf,    NULL);
  gss_display_name (&minor, cert_info->owner_name, &owner_name_buf, NULL);

  /* Some entries will be the same in both certificates.  Display them once */
  if (enc)
  {
    printGSSBufferDesc (pExitLocalVars, pre_s, "CA is", ca_name_buf);
    printGSSBufferDesc (pExitLocalVars, pre_s, "Certificate owner is", owner_name_buf);
    printGSSBufferDesc (pExitLocalVars, pre_s, "Certificates were created by Entrust Manager version",
	cert_info->manager_version);
    printGSSBufferDesc (pExitLocalVars, pre_s, "CRL distribution point is", cert_info->dist_point);
    printGSSBufferDesc (pExitLocalVars, pre_s, "Email address is", cert_info->email_addr);
  }

  gss_release_buffer (&minor, &ca_name_buf);
  gss_release_buffer (&minor, &owner_name_buf);

  dtkDebug (pExitLocalVars, "%s: %s certificate, version=V%d, serial number=%lu\n", pre_s,
    enc ? "Encryption  " : "Verification", cert_info->version+1, cert_info->serial_no);

  /* issue and expiry dates */
  dtkDebug2 (pExitLocalVars, "%s: Certificate issued  %s", pre_s, ctime(&cert_info->issue_date));
  dtkDebug2 (pExitLocalVars, "%s: Certificate expires %s", pre_s, ctime(&cert_info->expiry_date));
  dtkDebug2 (pExitLocalVars, "%s: Private %s key expires %s", pre_s, enc ? "decryption" : "signing",
	ctime(&cert_info->priv_key_end_time));
  if (!cert_info->rollover_allowed)
    dtkDebug2 (pExitLocalVars, "%s: Key update is not allowed\n", pre_s);

  /* display public key */
  dtkDebug2 (pExitLocalVars, "%s: %d bit PKCS-1 formatted public %s key...\n", pre_s,
      cert_info->public_key_length , enc ? "encryption" : "verification");
  printBinaryBuffer (pExitLocalVars, pre_s, 48, cert_info->public_key);

  /* release the certificate info buffer */
  ents_release_certificate_info (&minor, &cert_info) ;
}


/**********************************************************************/
/*                                                                    */
/* Function Name: displayCredCertificates                             */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*       Called to display information in the certificates associated */
/*       with user's credentials                                      */
/*                                                                    */
/* Input Parameters:  pExitLocalVars    - Local variables             */
/*                    pre_s             - Display message prefix      */
/*                    cred              - Security credentials        */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns:           None                                            */
/*                                                                    */
/**********************************************************************/

static void displayCredCertificates (PEXITLOCALVARS pExitLocalVars,
	           char *pre_s,
	           gss_cred_id_t cred)
{
  OM_uint32 status, minor;
  ents_cert_id_t cert;

  /* Encryption certificate */
  status = ents_get_ecertificate (&minor, cred, &cert);

  if (status == GSS_S_COMPLETE)
  {
    displayCertificate (pExitLocalVars, pre_s, cert, 1);
    ents_release_certificate (&minor, &cert);
  }

  /* Verification certificate */
  status = ents_get_vcertificate (&minor, cred, &cert);

  if (status == GSS_S_COMPLETE)
  {
    displayCertificate (pExitLocalVars, pre_s, cert, 0);
    ents_release_certificate (&minor, &cert);
  }

}


/**********************************************************************/
/*                                                                    */
/* Function Name: displayPeerCertificates                             */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*       Called to display information in the certificates of the     */
/*       context peer                                                 */
/*                                                                    */
/* Input Parameters:  pExitLocalVars    - Local variables             */
/*                    pre_s             - Display message prefix      */
/*                    ctx               - Security context            */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: None                                                      */
/*                                                                    */
/**********************************************************************/

static void displayPeerCertificates (PEXITLOCALVARS pExitLocalVars,
	           char *pre_s,
	           gss_ctx_id_t ctxt)
{
  OM_uint32 ulMajor, ulMinor;
  ents_cert_id_t cert;

  /* Encryption certificate */
  ulMajor = ents_get_peer_ecertificate (&ulMinor, ctxt, &cert);
  if (IsStatusOK (pExitLocalVars, __FILE__, __LINE__, ulMajor, ulMinor,
    GSS_S_COMPLETE, GSS_S_G_NONE,
    "displayPeerCertificates()->displayPeerCertificates()", TRUE))

  {
    displayCertificate (pExitLocalVars, pre_s, cert, 1);
    ents_release_certificate (&ulMinor, &cert);
  }

  /* Verification certificate */
  ulMajor = ents_get_peer_vcertificate (&ulMinor, ctxt, &cert);
  if (IsStatusOK (pExitLocalVars, __FILE__, __LINE__, ulMajor, ulMinor,
    GSS_S_COMPLETE, GSS_S_G_NONE, NULL, TRUE))
  {
    displayCertificate (pExitLocalVars, pre_s, cert, 0);
    ents_release_certificate (&ulMinor, &cert);
  }
}


/*
 * Character strings that names the mechanism
 */

static char *oid_to_mech_str(gss_OID mech)
{
   static char spkm1[] = "SPKM-1";
   static char spkm2[] = "SPKM-2";
   static char unkn[] = "?";

   if (g_OID_equal(mech, gss_mech_spkm1))
      return spkm1;
   else if (g_OID_equal(mech, gss_mech_spkm2))
      return spkm2;
   else
      return unkn;
}


/**********************************************************************/
/*                                                                    */
/* Function Name: displaySecContext                                   */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*       Called to display information that characterizes a           */
/*       security context                                             */
/*                                                                    */
/* Input Parameters:  pExitLocalVars    - Local variables             */
/*                    pre_s             - Display message prefix      */
/*                    ctx               - Security context            */
/*                    display_peer_certs- Display peer's certificat   */
/*                                        flag                        */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns:                       None                                                        */
/*                                                                    */
/**********************************************************************/

void displaySecContext (PEXITLOCALVARS pExitLocalVars,
	    char *pre_s,
	    gss_ctx_id_t ctxt,
	    int display_peer_certs)
{
  OM_uint32 ulMajor, ulMinor;
  gss_name_t src_name, targ_name, peer_name;
  gss_buffer_desc name_buf;
  OM_uint32 lifetime;
  gss_OID mech;
  OM_uint32 ret_flags;
  int locally_initiated;
  gss_OID output_name_type;

  ulMajor = gss_inquire_context (&ulMinor,
	 ctxt,                   /* Context handle */
	 &src_name,              /* Initiator name */
	 &targ_name,             /* Target name */
	 &lifetime,              /* Context lifetime */
	 &mech,                  /* Security mechanism */
	 &ret_flags,             /* Context flags */
	 &locally_initiated,     /* Initiator flag */
	 NULL);                  /* Established flag (optional) */

  if (!IsStatusOK (pExitLocalVars, __FILE__, __LINE__, ulMajor, ulMinor,
      GSS_S_COMPLETE, GSS_S_G_NONE,
      "displayPeerCertificates()->gss_inquire_context()", TRUE))
  {
    return;
  }

  peer_name = locally_initiated ? targ_name : src_name;

  gss_display_name (&ulMinor, peer_name, &name_buf, &output_name_type);

  dtkDebug2 (pExitLocalVars, "%s: Security context %s %.*s\n", pre_s,
       locally_initiated ? "established to" : "accepted from",
       name_buf.length, name_buf.value);

  dtkDebug2 (pExitLocalVars, "%s: Mech=%s MUTUAL=%d REPL=%d SEQ=%d "
	"ANON=%d PROT=%d TRANS=%d Lifetime=%lu\n",
       pre_s,
       oid_to_mech_str(mech),
       ret_flags & GSS_C_MUTUAL_FLAG ? 1:0,
       ret_flags & GSS_C_REPLAY_FLAG ? 1:0,
       ret_flags & GSS_C_SEQUENCE_FLAG ? 1:0,
       ret_flags & GSS_C_ANON_FLAG ? 1:0,
       ret_flags & GSS_C_PROT_READY_FLAG ? 1:0,
       ret_flags & GSS_C_TRANS_FLAG ? 1:0,
       lifetime);

  /*
   * An application could query peer certificates at this point.
   */
  if (display_peer_certs)
  {
    if (g_OID_equal(gss_nt_anonymous, output_name_type))
    {
      /* peer is anonymous */
      dtkDebug (pExitLocalVars, "%s: Certificates of peer (%.*s) are not available\n",
	  pre_s, name_buf.length, name_buf.value);
    }
    else
    {
      dtkDebug2 (pExitLocalVars, "%s: Certificates of peer (%.*s)...\n",
	    pre_s, name_buf.length, name_buf.value);
      displayPeerCertificates (pExitLocalVars, pre_s, ctxt);
    }
  }

  gss_release_name (&ulMinor, &targ_name);
  gss_release_name (&ulMinor, &src_name);
  gss_release_buffer (&ulMinor, &name_buf);
}

#ifdef GSS_ENTRUST_API

/**********************************************************************/
/*                                                                    */
/* Function Name: initEntrustEnv                                      */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*             Called to initialize the Entrust environment           */
/*                                                                    */
/* Input Parameters:  pExitLocalVars                                  */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: MQLONG    return code value                               */
/*                                                                    */
/* Exit Normal: MQXCC_OK                                              */
/*                                                                    */
/* Exit Error:  MQXCC_CLOSE_CHANNEL      Initialization failure       */
/*                                                                    */
/**********************************************************************/

MQLONG initEntrustEnv (PEXITLOCALVARS pExitLocalVars)
{
  char pszPW[ENTRUST_PW_LEN+1];
  char szErrMsg[MAX_LINE+1];

  gss_buffer_desc gssBuffer;

  OM_uint32       uiMajor, uiMinor,
	iDisplayStatusFlag =0;

  dtkDebug (pExitLocalVars, "initEntrustEnv(): \n");
  assert (pExitLocalVars);

  if (getPassword (pExitLocalVars,
	   pExitLocalVars->szKeyTab,
	   pExitLocalVars->szLocalQMgrName,
	   pszPW, ENTRUST_PW_LEN) != 0)
  {
    dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> initEntrustEnv(): "
      "Could not retrieve <%s>'s password from keytab <%s>\n",
      __FILE__, __LINE__, pExitLocalVars->szLocalQMgrName,
      pExitLocalVars->szKeyTab);
    //return MQXCC_SUPPRESS_FUNCTION;
    return MQXCC_CLOSE_CHANNEL; // Sever error no reason to retry.
  }

  // Import Entrust profile name
  gssBuffer.length = strlen (pExitLocalVars->szEntrustINIFile);
  gssBuffer.value  = pExitLocalVars->szEntrustINIFile;

  dtkDebug (pExitLocalVars, "initEntrustEnv(): Calling \"ents_set_inifile()\"\n");

  sprintf (szErrMsg, "initEntrustEnv()->ents_set_inifile() for file <%s>",
       pExitLocalVars->szEntrustINIFile);

  uiMajor = ents_set_inifile (&uiMinor, &gssBuffer);

  if (!IsStatusOK (pExitLocalVars, __FILE__, __LINE__, uiMajor, uiMinor,
    GSS_S_COMPLETE, GSS_S_G_NONE, szErrMsg, TRUE))
  {
    return MQXCC_SUPPRESS_FUNCTION;
  }

  gssBuffer.length = strlen (pszPW);
  gssBuffer.value  = pszPW;

  dtkDebug (pExitLocalVars, "initEntrustEnv(): Calling \"ents_set_password()\"\n");

  uiMajor = ents_set_password (&uiMinor, &gssBuffer);
  if (!IsStatusOK (pExitLocalVars, __FILE__, __LINE__, uiMajor, uiMinor,
    GSS_S_COMPLETE, GSS_S_G_NONE, "initEntrustEnv()->ents_set_password()", TRUE))
  {
    memset (pszPW, 0, ENTRUST_PW_LEN);
    return MQXCC_SUPPRESS_FUNCTION;
  }

  memset (pszPW, 0, ENTRUST_PW_LEN);

  return MQXCC_OK;
}

#endif  // GSS_ENTRUST_API


/**********************************************************************/
/*                                                                    */
/* Function Name: compareMQNames                                      */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*             Called to compare the                                  */
/*             target partner name which may be present in            */
/*             the ExitData of the                                    */
/*             channel definition with the PartnerName field          */
/*             of the channel exit parameters.                        */
/*                                                                    */
/* Input Parameters:  pExitLocalVars                                  */
/*                    pExitData                                       */
/*                    pPartnerName                                    */
/*                    ChannelType                                     */
/*                    pChannelName                                    */
/*                    pThisQMgrName                                   */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: MQLONG    return code value                               */
/*                                                                    */
/* Exit Normal: MQXCC_OK                                              */
/*                                                                    */
/* Exit Error:  MQXCC_SUPPRESS_FUNCTION     names don't match         */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Notes:                                                             */
/*   Restrictions: List any restrictions (or None)                    */
/*                                                                    */
/**********************************************************************/

MQLONG  compareMQNames (PEXITLOCALVARS  pExitLocalVars,
	    PMQCHAR                 pTargetName,
	    PMQCHAR                 pPartnerName,
	    MQLONG                  ChannelType,
	    PMQCHAR                 pChannelName,
	    PMQCHAR                 pThisQMgrName)
{
  int i;
  char szThisQMgrName[MQ_Q_MGR_NAME_LENGTH+1];
  char szChannelName[MQ_MCA_NAME_LENGTH+1];
  char szPartnerName[MQ_Q_MGR_NAME_LENGTH+1];
  char szTargetName[MQ_Q_MGR_NAME_LENGTH+1];

  convertMQCHARToString (szThisQMgrName, (PMQCHAR)pThisQMgrName, MQ_Q_MGR_NAME_LENGTH);
  convertMQCHARToString (szChannelName,  pChannelName,  MQ_MCA_NAME_LENGTH);
  convertMQCHARToString (szPartnerName,  pPartnerName,  MQ_Q_MGR_NAME_LENGTH);
  convertMQCHARToString (szTargetName,   pTargetName,   MQ_Q_MGR_NAME_LENGTH);

  dtkDebug (pExitLocalVars, "compareMQNames(): szThisQMgrName <%s>, szChannelName <%s> "
      "connected to szPartnerName <%s>, expected szTargetName <%s>\n",
      szThisQMgrName, szChannelName, szPartnerName, szTargetName);

  if (ChannelType == MQCHT_SVRCONN)
    return MQXCC_OK; /* don't know PartnerName for remote MQ clients */

  if ((szTargetName[0] == ' ') || (szTargetName[0] == '\0'))
    return MQXCC_OK; /* no target name */

  for (i = 0; i < MQ_EXIT_DATA_LENGTH; i++)
  {
    /* If all the non-space characters at the starts of the target names */
    /* have matched the check is successful.                             */

    if ((szTargetName[i] == ' ') && (szPartnerName[i] == ' ') ||
      (szTargetName[i] == '\0') && (szPartnerName[i] == '\0'))
      return MQXCC_OK;

    if (szTargetName[i] != szPartnerName[i])
    {       /* mismatch of names - May need to add strip blanks to another location */
      dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> compareMQNames(): "
	"-- SECURITY VIOLATION -- Local QMgr <%s>'s channel <%s> has detected an unauthorized "
	"request to connect from remote QMgr <%s>.  Local QMgr <%s> is "
	"expecting connection from QMgr <%s>.\n",
	__FILE__, __LINE__, szThisQMgrName, szChannelName, szPartnerName,
	szThisQMgrName, szTargetName);
      return MQXCC_SUPPRESS_FUNCTION;
    }

  } /* end for loop */

  return MQXCC_OK; /* full-length correct target name */

}


/**********************************************************************/
/*                                                                    */
/* Function Name: compareDNNames                                      */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*             Called to compare the                                  */
/*             target partner distinguished name (DN).  The valued    */
/*             expected, is found in the Channel Exit INI file.       */
/*                                                                    */
/* Input Parameters:  pExitLocalVars                                  */
/*                    pszTargetNameDN                                 */
/*                    PartnerGSSName                                  */
/*                    pChannelName                                    */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: MQLONG    return code value                               */
/*                                                                    */
/* Exit Normal: MQXCC_OK                                              */
/*                                                                    */
/* Exit Error:  MQXCC_SUPPRESS_FUNCTION     names don't match         */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Notes:                                                             */
/*   Restrictions: List any restrictions (or None)                    */
/*                                                                    */
/**********************************************************************/

MQLONG  compareDNNames (PEXITLOCALVARS  pExitLocalVars,
	    char*                   pszTargetNameDN,
	    gss_name_t              PartnerGSSName,
	    PMQCHAR                 pChannelName)
{
  char    szChannelName[MQ_MCA_NAME_LENGTH+1];
  char*   pszPartnerNameDN = NULL;

  if (PartnerGSSName != NULL)
  {
    pszPartnerNameDN = convertGSSNameToString (pExitLocalVars, PartnerGSSName);
    if (pszPartnerNameDN)
    {
      convertMQCHARToString (szChannelName,  pChannelName,  MQ_MCA_NAME_LENGTH);

      dtkDebug (pExitLocalVars, "compareDNNames(): szThisQMgrName <%s>, szChannelName <%s> "
	"connected to pszPartnerNameDN <%s>, expected pszTargetNameDN <%s>\n",
	pExitLocalVars->szLocalQMgrName, szChannelName, pszPartnerNameDN, pszTargetNameDN);

      if (strcmp (pszPartnerNameDN, pExitLocalVars->pszTargetNameDN) != 0)
      {
	dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> acceptSecContext(): "
	  "-- SECURITY VIOLATION -- Partner Distinguished Name (DN) <%s> of partner "
	  "QMgr <%s> is not the same as expected Target Distinguished Name (DN) "
	  "<%s> specified in file <%s>\n",
	  __FILE__, __LINE__, pszPartnerNameDN, pExitLocalVars->mqTargetName,
	    pExitLocalVars->pszTargetNameDN, pExitLocalVars->szChannelExitINIFile);

	dtkDebug (pExitLocalVars, "compareDNNames(): Freeing pszPartnerNameDN of size "
	  "<%ld>\n", strlen (pszPartnerNameDN) + 1);

	free (pszPartnerNameDN);
	return MQXCC_SUPPRESS_FUNCTION;
      }

      dtkDebug (pExitLocalVars, "compareDNNames(): Freeing pszPartnerNameDN of size "
	"<%ld>\n", strlen (pszPartnerNameDN) + 1);

      free (pszPartnerNameDN);
      return MQXCC_OK;
    }
  }

  dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> acceptSecContext(): "
	    "No Partner Distinguished Name (DN) in  found in "
	    "pSecData->PartnerGSSName - Aborting...\n",
	    __FILE__, __LINE__);

  return MQXCC_SUPPRESS_FUNCTION;
}


/**********************************************************************/
/*                                                                    */
/* Function Name: initSecurityExit                                    */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*           Called to initialize:                                    */
/*                                                                    */
/*             The SECEXITDATA in the ExitUserArea                    */
/*                                                                    */
/*             The CONTEXTDATA in the ExitData                        */
/*                                                                    */
/* Input Parameters:  pExitData                                       */
/*                    pChannelName                                    */
/*                                                                    */
/* Output Parameters: pSecData                                        */
/*                    pContext                                        */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: MQLONG    return code value                               */
/*                                                                    */
/* Exit Normal: MQXCC_OK                                              */
/*                                                                    */
/* Exit Error:  MQXCC_SUPPRESS_FUNCTION     unable to allocate memory */
/*                                                                    */
/**********************************************************************/

MQLONG  initSecurityExit (PMQCHAR32     pExitData,      // pParam passed from msgdata()
	      char*                 szChannelName,
	      PEXITUSERAREA pUserExitArea,
	PCONTEXTDATA  pContext)
{
  PSECEXITDATA    pSecData;
  PEXITLOCALVARS  pExitLocalVars;
  char szINIFile[MQ_EXIT_DATA_LENGTH+1];
  //char szChannelName[MQ_MCA_NAME_LENGTH+1];

  convertMQCHARToString (szINIFile, (PMQCHAR)pExitData, MQ_EXIT_DATA_LENGTH);
  //convertMQCHARToString (szChannelName, (PMQCHAR)pChannelName, MQ_MCA_NAME_LENGTH);

#ifdef DEBUG
  dtkDebug (NULL, "initSecurityExit(): INI file <%s>, ChannelName <%s>\n",
	szINIFile[0] ? szINIFile : "null", szChannelName);
#endif

  dtkDebug (NULL, "initSecurityExit(): Allocated pSecExitData for "
	"<%ld> bytes\n", sizeof (SECEXITDATA));

  pUserExitArea->pSecExitData = calloc (sizeof (SECEXITDATA), 1);
  pSecData = pUserExitArea->pSecExitData;

  // readINIFile() will allocate memory for pUserExitArea->pExitLocalVars
  // and initialize variables.  readINIFile() also ensures that
  // the Queue Manager Target is set
  if (readINIFile (szINIFile, &pUserExitArea->pExitLocalVars) != 0)
  {
    return MQXCC_SUPPRESS_FUNCTION;
  }

  pExitLocalVars = pUserExitArea->pExitLocalVars;

  if (pExitLocalVars->cDebug != EAMS_DEBUG_NONE)
    dumpINIFields (pExitLocalVars, szINIFile);

  // At this point the mqTargetName may be blank for SVRCONN
  // channel type.
  strcpy (pSecData->mqTargetName, pExitLocalVars->mqTargetName);

  pSecData->expectedSecMsg   = GSS_SEC_NONE;

  pSecData->PartnerGSSName   = NULL;
  pSecData->credentialHandle = (gss_cred_id_t)NULL;

  memset (pContext, 0, sizeof(CONTEXTDATA));

  return MQXCC_OK;

}

/*
 *
 */
MQLONG  initMessageExit (PMQCHAR32 pExitData,  // pParam passed from msgdata()
	     char*                  szChannelName,
	     PEXITUSERAREA  pUserExitArea)
{
  PEXITLOCALVARS  pExitLocalVars;
  char szINIFile[MQ_EXIT_DATA_LENGTH+1];

  convertMQCHARToString (szINIFile, (PMQCHAR)pExitData, MQ_EXIT_DATA_LENGTH);


  dtkDebug (NULL, "initMessageExit(): INI file <%s>, ChannelName <%s>\n",
	szINIFile[0] ? szINIFile : "null", szChannelName);

  // readINIFile() will allocate memory for pUserExitArea->pExitLocalVars
  // and initialize variables.  readINIFile() also ensures that
  // the Queue Manager Target is also set
  if (readINIFile (szINIFile, &pUserExitArea->pExitLocalVars) != 0)
  {
    return MQXCC_SUPPRESS_FUNCTION;
  }

  pExitLocalVars = pUserExitArea->pExitLocalVars;

  if (pExitLocalVars->cDebug != EAMS_DEBUG_NONE)
    dumpINIFields (pExitLocalVars, szINIFile);

  return MQXCC_OK;
}

/**********************************************************************/
/*                                                                    */
/* Function Name: savePartnerName                                     */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*          Converts a printable name into a GSS internal name        */
/*          format.                                                   */
/*          Uses the call gss_import_name.                            */
/*                                                                    */
/* Input Parameter:   pPartnerName                                    */
/*                    pChannelName                                    */
/*                                                                    */
/* Output Parameters: *pPartnerGSSName                                */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: MQLONG    return code value                               */
/*                                                                    */
/**********************************************************************/

MQLONG savePartnerName (PEXITUSERAREA   pExitUserArea,
	    gss_name_t              *pPartnerGSSName,
	    PMQCHAR48               pPartnerName,
	    PMQBYTE                 pChannelName)
{
  char szPartnerName[MQ_Q_MGR_NAME_LENGTH+1];
  char szChannelName[MQ_MCA_NAME_LENGTH+1];

  PEXITLOCALVARS pExitLocalVars = pExitUserArea->pExitLocalVars;

  convertMQCHARToString (szPartnerName, (PMQCHAR)pPartnerName, MQ_Q_MGR_NAME_LENGTH);
  convertMQCHARToString (szChannelName, (PMQCHAR)pChannelName, MQ_MCA_NAME_LENGTH);

  dtkDebug (pExitLocalVars, "savePartnerName(): pPartnerName <%s>, pChannelName <%s>\n",
    szPartnerName[0] ? szPartnerName : "null", szChannelName);

  if ((*pPartnerName[0] != '\0') && (*pPartnerName[0] != ' '))
  {
    OM_uint32       major_status = 0;
    OM_uint32       minor_status = 0;

    #if MQAT==MQAT_OS2
    #pragma pack(4)
    gss_buffer_desc name;
    #pragma pack(1)
    #else
    gss_buffer_desc name;
    #endif

#ifdef GSS_ENTRUST_API
    name.length = strlen (pExitLocalVars->pszTargetNameDN) + 1;
    name.value  = pExitLocalVars->pszTargetNameDN;
#else
    char szQmNameStr[MQ_Q_MGR_NAME_LENGTH+1];

    memset (szQmNameStr, 0, MQ_Q_MGR_NAME_LENGTH + 1);

    strncpy (szQmNameStr, (char*)pPartnerName, MQ_Q_MGR_NAME_LENGTH);
    strtok (szQmNameStr, " " ); /* null terminate QM name */

    name.length = strlen (szQmNameStr) + 1;
    name.value  = szQmNameStr;
#endif  // GSS_ENTRUST_API

    major_status = gss_import_name (&minor_status,
	            &name,
	            GSS_C_NULL_OID,
	            pPartnerGSSName );

    if (IsStatusOK (pExitLocalVars, __FILE__, __LINE__, major_status, minor_status,
      GSS_S_COMPLETE, GSS_S_G_NONE, "savePartnerName()->gss_import_name()", TRUE))
    {
      return MQXCC_OK;
    }
    else
    {
       /* gss_import_name error */
      dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> savePartnerName(): "
	"Unexpected error from <gss_import_name()>\n",
	__FILE__, __LINE__);

      return MQXCC_SUPPRESS_FUNCTION;
    }

  }
  else
  {       /* No PartnerName */
    dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> savePartnerName(): "
	"No pPartnerName supplied\n", __FILE__, __LINE__);
    return MQXCC_SUPPRESS_FUNCTION;
  }
}


/**********************************************************************/
/*                                                                    */
/* Function Name: acquireCred                                         */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*          Acquire credentials by:                                   */
/*                                                                    */
/*          Convert the principal into a GSS internal format          */
/*                                                                    */
/*          Call gss_acquireCred so the principal can initiate and    */
/*          accept security contexts.                                 */
/*                                                                    */
/* Input Parameters:  pChDef                                          */
/*                    pName                                           */
/*                                                                    */
/* Output Parameters: *pCredHandle                                    */
/*                    *pContextHandle                                 */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: MQLONG    return code value                               */
/*                                                                    */
/**********************************************************************/

MQLONG acquireCred (PMQCXP                pParms,
	  PMQCD         pChDef,
	  gss_cred_id_t *pCredHandle,
	gss_ctx_id_t  *pContextHandle,
	PMQBYTE       pName)
{
  OM_uint32       major_status = 0;
  OM_uint32       minor_status = 0;
  OM_uint32       extra_minor_status = 0;
  gss_OID_set     actual_mechs = NULL;
  OM_uint32       actual_time;
  gss_name_t      desired_name = NULL;
  gss_buffer_desc name;
  gss_OID                 gssNameType;

  PEXITUSERAREA   pExitUserArea  = (PEXITUSERAREA)&(pParms->ExitUserArea);
  PEXITLOCALVARS  pExitLocalVars = pExitUserArea->pExitLocalVars;

#ifdef GSS_DCE_API
  char            FullPathObjName[512];
#endif

  *pContextHandle = GSS_C_NO_CONTEXT;
  if (pName == NULL)
  { /* no name: default credential on gss_init_sec_context() */
    *pCredHandle = GSS_C_NO_CREDENTIAL;
    return MQXCC_OK;
  }

  /*******************************************************************/
  /* We have been given a name: convert it into GSS                  */
  /* format                                                          */
  /*******************************************************************/
  name.length = strlen((char *)pName)+1;
  name.value  = pName;

  if (strstr ((char*) name.value, "cn=") != NULL)
  //if (strstr ((char*) name.value, "cn=") == (char*) name.value)
  gssNameType = gss_nt_ldap_dn;
    else
    gssNameType = GSS_SPKM_NT_USER_NAME;

  dtkDebug (pExitLocalVars, "acquireCred(): Calling \"gss_import_name()\"\n");

  major_status = gss_import_name (&minor_status,
	          &name,
	          gssNameType,
	          &desired_name);

  if (!IsStatusOK (pExitLocalVars, __FILE__, __LINE__, major_status, minor_status,
    GSS_S_FAILURE, GSS_S_G_OFFLINE_MODE, "acquireCred()->gss_import_name()", TRUE))
  {
    return MQXCC_SUPPRESS_FUNCTION;
  }

#ifdef GSS_DCE_API
  /*******************************************************************/
  /* Call gssdce_register_acceptor_identity so the subsequent        */
  /* gss_acquire_cred call uses the keytable with the principal's    */
  /* name.                                                           */
  /*******************************************************************/

  #if MQAT_DEFAULT==MQAT_OS2 || MQAT_DEFAULT==MQAT_WINDOWS_NT

  /* Set FullPathObjName to the workpath, ie the directory         */
  /* in which mqs.ini lives.                                       */
  /* For example: strcpy(FullPathObjName,"c:\\mqm");               */

  /* Complete the string defining the full path of the keytable */
  strcat (FullPathObjName,"\\dce\\keytabs\\");

  #endif

  #if MQAT_DEFAULT==MQAT_UNIX

  /* UNIX: Setup the string defining the full path of the keytable */
  strcpy (FullPathObjName,"/var/mqm/dce/keytabs/");

  #endif

  strcat (FullPathObjName,(const char*)pName);


  major_status = gssdce_register_acceptor_identity (&minor_status,
	                    desired_name,
	                    NULL,
	                    FullPathObjName);


  if (major_status != GSS_S_COMPLETE)
  {                       /* DCE call failed */

    /* Diagnostic output here */

    gss_release_name (&minor_status, &desired_name);
    return MQXCC_SUPPRESS_FUNCTION;

  }
#endif

  /*******************************************************************/
  /* Call gss_acquire_cred so the principal can initiate and accept  */
  /* security contexts under the 'desired_name' identity.            */
  /*******************************************************************/
  dtkDebug (pExitLocalVars, "acquireCred(): Calling \"gss_acquire_cred()\"\n");

  major_status = gss_acquire_cred (&minor_status, desired_name,
	          0, GSS_C_NULL_OID_SET, GSS_C_BOTH,
	          pCredHandle,
	          &actual_mechs, &actual_time);
  if (IsStatusOK (pExitLocalVars, __FILE__, __LINE__, major_status, minor_status,
	  GSS_S_FAILURE, GSS_S_G_OFFLINE_MODE,
	  "acquireCred()->gss_acquire_cred()", TRUE))
  {
    if (pExitLocalVars->cDebug & (EAMS_DEBUG_CONSOLE | EAMS_DEBUG_FILE))
      displayCredCertificates (pExitLocalVars, baseName ((char*)pName), *pCredHandle);
    return MQXCC_OK;
  }

  /**********************************************************/
  /* For an MQ client we tried above to acquireCred         */
  /* for the login name of the user who started the client  */
  /* application.                                           */
  /* This works if this login name is also a                */
  /* principal which is defined to DCE.                     */
  /*                                                        */
  /* It is also OK if it fails with a no cred               */
  /* response. In this case we know that the                */
  /* principal is not defined to DCE. We set pCredHandle    */
  /* such that gss_init_sec_context will attempt to use     */
  /* the default DCE credential.                            */
  /*                                                        */
  /* MQ server principals must be defined to DCE.           */
  /**********************************************************/
  if ((pChDef->ChannelType == MQCHT_CLNTCONN) &&
    (major_status == GSS_S_NO_CRED) )
  {
    *pCredHandle = GSS_C_NO_CREDENTIAL;
    return MQXCC_OK;
  }

  /* The gss_acquire_cred call failed */

  /* Diagnostic output here */

  major_status = gss_release_name (&minor_status, &desired_name);
  //return MQXCC_SUPPRESS_FUNCTION;       // Was in sample
  return MQXCC_CLOSE_CHANNEL; // Sever error no reason to retry.
}



/**********************************************************************/
/*                                                                    */
/* Function Name: createCreds                                         */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*          Create the credentials                                    */
/*                                                                    */
/* Input Parameters: pChDef                                           */
/*                                                                    */
/* Output Parameters:*pCredHandle                                     */
/*                   *pContextHandle                                  */
/*                                                                    */
/* Returns: MQLONG    return code value                               */
/*                                                                    */
/**********************************************************************/

MQLONG createCreds (PMQCXP                      pParms,
	  PMQCD                   pChDef,
	  gss_cred_id_t   *pCredHandle,
	  gss_ctx_id_t    *pContextHandle)
{
  MQLONG rc;
  PMQBYTE pClientLoginName;

  /**********************************************************************/
  /* This sectiona allows for the Windows NT equivalent for getting the */
  /* logged user.                                                       */
  /**********************************************************************/
  #if MQAT_DEFAULT==MQAT_WINDOWS_NT
   #include <windows.h>             /* Windows standard header         */
   #include <winbase.h>             /* GetUserName...                  */
   #define MAX_NAME_LEN 256         /* Maximum User Name length        */
   LPTSTR lpBuffer = (LPTSTR)NULL;  /* Input buffer for Logged on User */
   DWORD  dwSize   = MAX_NAME_LEN+1;/* Maximum buffer size             */
   BOOL   bSuccess = FALSE;         /* Return from GetUserName         */
   BOOL   bRetry   = FALSE;         /* Retry GetUserName call?         */
   LONG   NumRetries = 0;           /* Number of retries so far        */
   DWORD  error;
  #endif /* MQAT_WINDOWS_NT */
  #if MQAT_DEFAULT==MQAT_UNIX
   #ifndef LOGNAME_MAX
     #define LOGNAME_MAX 8          /* Maximum logon name length       */
   #endif
   int    getlogin_r_rc;            /* Return value from getlogin_r    */
   size_t nameLen = LOGNAME_MAX+1;  /* Length buffer for login name    */
  #endif

  PEXITLOCALVARS pExitLocalVars = ((PEXITUSERAREA)pParms->ExitUserArea)->pExitLocalVars;
  dtkDebug (pExitLocalVars, "createCreds():\n");

  if (pChDef->ChannelType == MQCHT_CLNTCONN)
  {
    /*****************************************************************/
    /* For an exit running as part of an MQ client, if appopriate for*/
    /* the platform we find out the name that the user who started   */
    /* that client application used when he logged in to the client  */
    /* operating system.                                             */
    /* If this user is also defined as a principal to DCE this       */
    /* is the principal which will be used.                          */
    /* Otherwise the client exit uses the default principal.         */
    /*****************************************************************/
    #if MQAT_DEFAULT==MQAT_WINDOWS_NT || MQAT_DEFAULT==MQAT_UNIX
    /* Let's get some memory for the buffer */
    #if MQAT_DEFAULT==MQAT_WINDOWS_NT
    dtkDebug (pExitLocalVars, "createCreds(): Allocating lpBuffer for "
	"<%ld> bytes\n", MAX_NAME_LEN + 1);

    lpBuffer = (LPTSTR) malloc (MAX_NAME_LEN + 1);
    if (lpBuffer == NULL)
    #elif MQAT_DEFAULT==MQAT_UNIX
    dtkDebug (pExitLocalVars, "createCreds(): Allocating pClientLoginName "
      "for <%ld> bytes\n", LOGNAME_MAX + 1);
    pClientLoginName = (PMQBYTE) malloc (LOGNAME_MAX + 1);
    if (pClientLoginName == NULL)
    #endif
    {
      /* We failed to allocate memory */
      /* Diagnostic output here */
      return MQXCC_SUPPRESS_FUNCTION;
    }
    #endif /* MQAT_DEFAULT==MQAT_WINDOWS_NT || MQAT_DEFAULT==MQAT_UNIX */

    #if MQAT_DEFAULT==MQAT_WINDOWS_NT

    /**************************************************************/
    /* Get the NT login user name                                 */
    /**************************************************************/
    for (bRetry = TRUE, NumRetries = 60;
      (bRetry) && (NumRetries > 0); NumRetries--)
    {
      bRetry = FALSE;
      bSuccess = GetUserName (lpBuffer, &dwSize);
      if (!bSuccess)
      {
	error = GetLastError ();
	/**************************************************************/
	/* If Server is busy, wait a second & retry                   */
	/**************************************************************/
	if ((error == RPC_S_SERVER_TOO_BUSY)
	  ||(error == RPC_S_OUT_OF_RESOURCES)
	  ||(error == RPC_S_PROTOCOL_ERROR)
	  ||(error == RPC_S_CALL_FAILED))
	{
	  bRetry = TRUE;
#if MQAT_DEFAULT==MQAT_UNIX
	  sleep (1);
#else
	  Sleep(1000L);
#endif
	}
      }
    }
    if (bSuccess)     /* done looping: it either worked or it didn't */
    {
    /* Return was fine, copy the buffer across */
      pClientLoginName = (PMQBYTE) lpBuffer;
    }
    else
    {
      /* NULL if can't get login name, explicitly set the value     */
      /* to NULL: indicates should try to use DCE default login.    */
      pClientLoginName = (PMQBYTE) NULL;

      /* Diagnostic output here */

    }
    #elif MQAT_DEFAULT==MQAT_OS2
    pClientLoginName = NULL; /* no login name */
    #else

    /**************************************************************/
    /* Get the UNIX login user name                               */
    /**************************************************************/

    /* use getlogin_r as exit can be called on each MQCONN on a threaded client */
    getlogin_r_rc = (int)getlogin_r ((char *)pClientLoginName,nameLen); /* get login name */
    if (getlogin_r_rc != 0)
    {
      /* getlogin_r call failed */

      /* NULL if can't get login name, explicitly set the value     */
      /* to NULL: indicates should try to use DCE default login.    */
      pClientLoginName = (PMQBYTE) NULL;

      /* The location of the error value depends on the POSIX level of the UNIX platform. */
      /* Early POSIX versions return -1 and provide the error value in errno;             */
      /* later ones provide the error value on the function call return.                  */
      /* As these returned error codes are never -1, can test getlogin_r_rc and if it is  */
      /* -1 can assume that we are desling with an early POSIX system.                    */

      if (getlogin_r_rc == -1) getlogin_r_rc = errno;

      /* now getlogin_r_rc has the error value whichever Posix standard is involved.      */

      /* Diagnostic output here */

    }

    #endif

    rc = acquireCred (pParms, pChDef, pCredHandle, pContextHandle, pClientLoginName);

    #if MQAT_DEFAULT==MQAT_WINDOWS_NT
    if (lpBuffer)
    {
      dtkDebug (pExitLocalVars, "createCreds(): Freeing lpBuffer of size "
	"<%ld>\n", MAX_NAME_LEN + 1);
      free (lpBuffer);
    }
    #elif MQAT_DEFAULT==MQAT_UNIX
    if (pClientLoginName)
    {
      dtkDebug (pExitLocalVars, "createCreds(): Freeing pClientLoginName\n");
      free (pClientLoginName);
    }
    #endif
  }
  else
  {
    /*****************************************************************/
    /* If we're not on a client find out the name of the Q Manager   */
    /* the Q Manager name is held in pChDef -> QMgrName              */
    /*****************************************************************/

    /*****************************************************************/
    /* Having got the queue manager name use it to acquire the GSS   */
    /* credential.                                                   */
    /*****************************************************************/
    rc = acquireCred (pParms,
	      pChDef,
	      pCredHandle,
	      pContextHandle,
#ifdef GSS_ENTRUST_API
	      (PMQBYTE) pExitLocalVars->szEntrustEPFFile);
#elif  GSS_DCE_API
	      (PMQBYTE) pExitLocalVars->szLocalQMgrName);
#else
  #error No GSS_ENTRUST_API or GSS_DCE_API preprocessor defined
#endif

  } /* endif */

  return rc;
}

/*
 *  Set Key Establishment Protocol as specified in the INI file
 */
MQLONG setKeyEstblishmentProtocol (PEXITLOCALVARS pExitLocalVars,
	           gss_cred_id_t  credentialHandle)
{
  OM_uint32       ulMajor = 0;
  OM_uint32       ulMinor = 0;
  MQLONG rc;

  if (pExitLocalVars->cDebug != EAMS_DEBUG_NONE)
  {
    char* lpszBuff;

    switch (pExitLocalVars->entSymmetricKeyExchangeProtocol)
    {
      case SPKM_KEY_ESTB_PROTOCOL_RSA:
	lpszBuff = PROTOCOL_RSA_STRING;
	break;

      case SPKM_KEY_ESTB_PROTOCOL_D_H:
	lpszBuff = PROTOCOL_D_H_STRING;
	break;

      case SPKM_KEY_ESTB_PROTOCOL_DEFAULT:
	lpszBuff = PROTOCOL_DEFAULT_STRING;
	break;

      default:
	lpszBuff = "???????";
    }

    dtkDebug (pExitLocalVars, "setKeyEstblishmentProtocol(): "
	"Calling \"ents_set_key_estb_protocol\" using <%s> protocol\n",
	lpszBuff);
  }

  ulMajor = ents_set_key_estb_protocol (&ulMinor,
	        credentialHandle,
	        pExitLocalVars->entSymmetricKeyExchangeProtocol);
  if (IsStatusOK (pExitLocalVars, __FILE__, __LINE__, ulMajor, ulMinor,
    GSS_S_COMPLETE, GSS_S_G_NONE,
    "setKeyEstblishmentProtocol()->ents_set_key_estb_protocol()", TRUE))
    rc = MQXCC_OK;
  else    // else error
    rc = MQXCC_SUPPRESS_FUNCTION;

  return rc;
}

/**********************************************************************/
/*                                                                    */
/* Function Name: composeSecMsg                                       */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*          Construct a security message to send                      */
/*                                                                    */
/* Input Parameters: pHeader,                                         */
/*                   headerLength,                                    */
/*                   msgType,                                         */
/*                   msg,                                             */
/*                   msgLength,                                       */
/*                   exitReturnCode,                                  */
/*                   pChannelName.                                    */
/*                                                                    */
/* Output Parameters: pParms,                                         */
/*                    *pExitBufAddr,                                  */
/*                    pExitBufLen.                                    */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/*                                                                    */
/* Returns: None                                                      */
/*                                                                    */
/**********************************************************************/

void composeSecMsg (PMQCXP   pParms,
	PMQBYTE  *pExitBufAddr,
	PMQLONG  pExitBufLen,
	PMQBYTE  pHeader,
	MQLONG   headerLength, // 0 for authentication, > 0 during send/receive
	MQBYTE   msgType,      // 0    GSS_SEC_SEAL for send/receive
	PMQBYTE  msg,          // context token
	MQLONG   msgLength,    // context token length
	MQLONG   exitReturnCode,
	PMQCHAR  pChannelName)
{
  MQLONG       length;
  PMQBYTE      buffer;
  PGSSSECMSG   pGSS_secMsg;

  PEXITUSERAREA   pExitUserArea  = (PEXITUSERAREA)&(pParms->ExitUserArea);
  PEXITLOCALVARS  pExitLocalVars = pExitUserArea->pExitLocalVars;

  dtkDebug (pExitLocalVars, "composeSecMsg():\n");
  dtkDebug (pExitLocalVars, "composeSecMsg(): sizeof secmsg %d\n",sizeof(*pGSS_secMsg));
	               /* find the length of the data   */
  length = headerLength + sizeof (pGSS_secMsg->header) + msgLength;

  /* create a buffer to hold data  */
  dtkDebug (pExitLocalVars, "composeSecMsg(): Allocating lpBuffer for "
	"<%ld> bytes\n", length);
  buffer = (PMQBYTE) malloc ((size_t)length);

  if (buffer != NULL)
  {                                      /* make pGSS_secMsg = the buffer ptr*/
    pGSS_secMsg = (PGSSSECMSG)&buffer[headerLength];

    if ((pHeader != NULL) && (headerLength > 0))
    {                                  /* if there's a header copy to buffer*/
      memcpy (buffer, pHeader, (size_t)headerLength);
      /* if send exit set byte 9 to zero so receive exit knows it's been */
      /* encrypted */
      if (SENDEXITCALL) buffer[8] = '\0';
    } /* endif */
#if MQAT_DEFAULT==MQAT_WINDOWS_NT
  #ifdef GSS_DCE_API
    pGSS_secMsg->header.mechanism = htons1(GSS_DCE_MACHANISM);
  #elif GSS_ENTRUST_API
    pGSS_secMsg->header.mechanism = htons1(GSS_ENTRUST_MECHANISM);
  #else
    #error No GSS_ENTRUST_API or GSS_DCE_API preprocessor defined
  #endif
#else
  #ifdef GSS_DCE_API
    pGSS_secMsg->header.mechanism = htons(GSS_DCE_MACHANISM);
  #elif GSS_ENTRUST_API
    pGSS_secMsg->header.mechanism = htons(GSS_ENTRUST_MECHANISM);
  #else
    #error No GSS_ENTRUST_API or GSS_DCE_API preprocessor defined
  #endif
#endif
    pGSS_secMsg->header.msgType = msgType;

    if ((msg != NULL) && (msgLength > 0))
    {                                     /* copy over the data            */
      memcpy (&(pGSS_secMsg->msg), msg, (size_t)msgLength);
    } /* endif */

    *pExitBufAddr = buffer;
    *pExitBufLen  = length;

    pParms->ExitResponse2 = MQXR2_USE_EXIT_BUFFER;
    pParms->ExitResponse = exitReturnCode;
  }
  else
  {                               /* unable to allocate memory       */
    /* Diagnostic output here */
    dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> composeSecMsg(): "
      "Unable to allocate memory\n\n", __FILE__, __LINE__);
    pParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
  } /* endif */
  /*
  {
    int i;
    dtkDebug2(pExitLocalVars,"MessageLength %0d\n",msgLength);
    for (i=0;i<length;i++)
    {
      dtkDebug2(pExitLocalVars,"%02X",buffer[i]);
      if (((i+1) % 8) == 0)
	dtkDebug2(pExitLocalVars,"\n");
    }
    dtkDebug2(pExitLocalVars,"\n");
  }
  */
}

/**********************************************************************/
/*                                                                    */
/* Function Name: composeSecTokenMsg                                  */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*          Compose a GSS token                                       */
/*                                                                    */
/* Input Parameters: major_status,                                    */
/*                   minor_status,                                    */
/*                   msgType,                                         */
/*                   *pOutputToken,                                   */
/*                   pChannelName.                                    */
/*                                                                    */
/* Output Parameters: pParms,                                         */
/*                    *pOutputMsgAddr,                                */
/*                    pOutputMsgLength.                               */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: None                                                      */
/*                                                                    */
/**********************************************************************/

void composeSecTokenMsg (PMQCXP          pParms,
	     OM_uint32       major_status,
	     OM_uint32       minor_status,
	     MQBYTE          msgType,
	     gss_buffer_desc *pOutputToken,
	     PMQBYTE         *pOutputMsgAddr,
	     PMQLONG         pOutputMsgLength,
	     PMQCHAR         pChannelName)
{
  PEXITUSERAREA   pUserExitArea;
  PEXITLOCALVARS  pExitLocalVars;

  pUserExitArea   = (PEXITUSERAREA)&(pParms->ExitUserArea);
  pExitLocalVars  = pUserExitArea->pExitLocalVars;

  dtkDebug (pExitLocalVars, "composeSecTokenMsg():\n");

  if (major_status == GSS_S_COMPLETE)
  {
    if (pOutputToken->length > 0)
    {
      composeSecMsg (pParms,
	       pOutputMsgAddr,
	       pOutputMsgLength,
	       NULL,
	       0,
	       msgType,
	       pOutputToken->value,
	       pOutputToken->length,
	       MQXCC_SEND_SEC_MSG,
	       pChannelName);
    }
    else                                        /* no token            */
    {
      dtkDisplay (pExitLocalVars, "composeSecTokenMsg(): gss_init_sec_context() succeeded "
	    "- Security Context Established\n");
      *pOutputMsgAddr   = NULL;
      *pOutputMsgLength = 0;
      pParms->ExitResponse = MQXCC_OK;
    } /* endif */
  }
  else
  {
    if (GSS_SUPPLEMENTARY_INFO(major_status) & GSS_S_CONTINUE_NEEDED)
    {
      if (pOutputToken->length > 0)
      {
	composeSecMsg (pParms,
	         pOutputMsgAddr,
	         pOutputMsgLength,
	         NULL,
	         0,
	         msgType,
	         pOutputToken->value,
	         pOutputToken->length,
	         MQXCC_SEND_AND_REQUEST_SEC_MSG,  // Ensure we are called
	         pChannelName);                   // back to complete operation
      }
    }
    else    /* no token     */
    {       /* gss_init_sec_context() call failed   */
      dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> composeSecTokenMsg(): "
	"Possible -- SECURITY VIOLATION -- Unexpected error from GSS-API call()\n",
	__FILE__, __LINE__);
      *pOutputMsgAddr   = NULL;
      *pOutputMsgLength = 0;

      pParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
    }
  }
}




/**********************************************************************/
/*                                                                    */
/* Function Name: initSecContext                                      */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*          Initiate Security context between this application and    */
/*          the partner                                               */
/*                                                                    */
/* Input Parameters: pSecData                                         */
/*                   pInputMsgAddr                                    */
/*                   inputMsgLength                                   */
/*                   pChannelName                                     */
/*                                                                    */
/* Output Parameters: pParms                                          */
/*                    *pOutputMsgAddr                                 */
/*                    pOutputMsgLength                                */
/*                                                                    */
/* InOut Parameters:  pContextData                                    */
/*                                                                    */
/* Returns: None                                                      */
/*                                                                    */
/**********************************************************************/

void initSecContext (PMQCXP       pParms,
	   PSECEXITDATA pSecData,
	   PCONTEXTDATA pContextData,
	   PMQBYTE      pInputMsgAddr,
	   MQLONG       inputMsgLength,
	 PMQBYTE      *pOutputMsgAddr,
	   PMQLONG      pOutputMsgLength,
	   PMQCHAR      pChannelName)
{
  gss_buffer_desc output_token = { 0, NULL };
  gss_OID         mech_type;              // GSS_C_NULL_OID == gss_mech_spkm1
  gss_OID         actual_mech_type;
  OM_uint32               req_flags    = GSS_C_MUTUAL_FLAG;
  OM_uint32               ret_flags;
  //OM_uint32       time_req     = 0;
  OM_uint32       time_rec     = 0;
  gss_buffer_desc input_token;
  gss_buffer_t    pInput_Token;
  OM_uint32       ulMajor = 0;
  OM_uint32       ulMinor = 0;

  PEXITUSERAREA   pUserExitArea;
  PEXITLOCALVARS  pExitLocalVars;

  pUserExitArea   = (PEXITUSERAREA)&(pParms->ExitUserArea);
  pExitLocalVars  = pUserExitArea->pExitLocalVars;

  dtkDebug (pExitLocalVars, "initSecContext():\n");

  mech_type = pExitLocalVars->entSecurityMechanism;

  /*
   * If this is the very first call initalize the input_token to GSS_C_NO_BUFFER
   */
  if (inputMsgLength == 0)
  {
    dtkDebug (pExitLocalVars, "initSecContext(): called for the very first time\n");
    pInput_Token = GSS_C_NO_BUFFER;
    pExitLocalVars->iInitSecContextCounter = 1;
  }
  else
  {
    ++(pExitLocalVars->iInitSecContextCounter);
#if 0
    if (pExitLocalVars->iInitSecContextCounter == 1)
      dtkDebug (pExitLocalVars, "initSecContext(): called for the very first time\n");
    else
#endif
      dtkDebug (pExitLocalVars, "initSecContext(): has been called <%d> times\n",
	    pExitLocalVars->iInitSecContextCounter);
    input_token.length = inputMsgLength;
    input_token.value  = pInputMsgAddr;
    pInput_Token = &input_token;
  }

  mech_type = pExitLocalVars->entSecurityMechanism;

  dtkDebug (pExitLocalVars, "initSecContext(): Calling \"gss_init_sec_context()\"\n");

  /*  initiate a security context  */
  ulMajor = gss_init_sec_context (&ulMinor,
	          pSecData->credentialHandle,
	          &(pContextData->contextHandle),
	          pSecData->PartnerGSSName,
	          mech_type, req_flags,
	          pExitLocalVars->ulReqContextLifeTime,
	          NULL, pInput_Token,
	          &actual_mech_type, &output_token,
	          &ret_flags, &time_rec);

  // If first time in cause a awarning to be displayed if call was
  // successful. Otherwise an error will be displayed.
  IsStatusOK (pExitLocalVars, NULL, 0, ulMajor, ulMinor,
    GSS_S_CONTINUE_NEEDED, GSS_S_G_NONE,
    "initSecContext()->gss_init_sec_context()", FALSE);

  if (pExitLocalVars->cDebug & (EAMS_DEBUG_FILE | EAMS_DEBUG_CONSOLE) &&
    (pExitLocalVars->iInitSecContextCounter == 1))
    displaySecContext (pExitLocalVars,
	       pExitLocalVars->szLocalQMgrName,
	       pContextData->contextHandle, 1);

  /* major_status is checked in composeSecTokenMsg */

  composeSecTokenMsg (pParms,
	    ulMajor,
	    ulMinor,
	    GSS_SEC_INIT_SEC,
	    &output_token,
	    pOutputMsgAddr,
	    pOutputMsgLength,
	    pChannelName);

  gss_release_buffer (&ulMinor, &output_token);
}


/**********************************************************************/
/*                                                                    */
/* Function Name: acceptSecContext                                    */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*           Establist a remotely initiated security context between  */
/*           this application and a remote Peer.                      */
/*                                                                    */
/* Input Parameters: pSecData,                                        */
/*                   pInputMsgAddr,                                   */
/*                   inputMsgLength.                                  */
/*                   pChannelName                                     */
/*                                                                    */
/* Output Parameters: pParms,                                         */
/*                    pOutputMsgAddr,                                 */
/*                    pOutputMsgLength.                               */
/*                                                                    */
/* InOut Parameters:  pContextData.                                   */
/*                                                                    */
/* Returns: None                                                      */
/*                                                                    */
/* Notes:                                                             */
/*   Restrictions: List any restrictions (or None)                    */
/*                                                                    */
/**********************************************************************/

void acceptSecContext (PMQCXP       pParms,
	   PSECEXITDATA pSecData,
	   PCONTEXTDATA pContextData,
	   PMQBYTE      pInputMsgAddr,
	   MQLONG       inputMsgLength,
	   PMQBYTE      *pOutputMsgAddr,
	   PMQLONG      pOutputMsgLength,
	   PMQCHAR      pChannelName)
{
  gss_buffer_desc output_token    = { 0, NULL };
  OM_uint32               context_support = 0;
  OM_uint32               context_time    = 0;
  gss_cred_id_t   delegated_cred_handle;
  OM_uint32               major_status    = 0;
  OM_uint32               minor_status    = 0;
  gss_OID                 mech_type       = GSS_C_NULL_OID;
  gss_buffer_desc input_token;

  PEXITUSERAREA   pExitUserArea   = (PEXITUSERAREA)&(pParms->ExitUserArea);
  PEXITLOCALVARS  pExitLocalVars  = pExitUserArea->pExitLocalVars;

  dtkDebug (pExitLocalVars, "acceptSecContext(): Calling "
	    "\"gss_accept_sec_context()\"\n");

  input_token.length = inputMsgLength;
  input_token.value  = pInputMsgAddr;

  mech_type = pExitLocalVars->entSecurityMechanism;

  major_status = gss_accept_sec_context (&minor_status,
	              &(pContextData->contextHandle),
	              pSecData->credentialHandle,
	              //&input_token, NULL,
	              &input_token, GSS_C_NO_CHANNEL_BINDINGS,
	              &(pSecData->PartnerGSSName),
	              &mech_type, &output_token, &context_support,
	              &context_time, &delegated_cred_handle);

  if (IsStatusOK (pExitLocalVars, __FILE__, __LINE__, major_status, minor_status,
      GSS_S_CONTINUE_NEEDED, GSS_S_G_NONE,
      "acceptSecContext()->gss_accept_sec_context()", FALSE))
  {
    pExitLocalVars->iAcceptSecContextCounter++;
    if (pExitLocalVars->iAcceptSecContextCounter == 1)
    {
      displaySecContext (pExitLocalVars,
	         pExitLocalVars->szLocalQMgrName,
	         pContextData->contextHandle,
	         1);

      dtkDebug (pExitLocalVars, "acceptSecContext(): called for the very first time\n");
    }
    else
    {
      dtkDebug (pExitLocalVars, "acceptSecContext(): has been called <%d> times\n",
	    pExitLocalVars->iAcceptSecContextCounter);
    }
  }

#ifdef GSS_ENTRUST_API
  if (pParms->ExitResponse != MQXCC_SUPPRESS_FUNCTION)
  {
    if (compareDNNames (pExitLocalVars,
	      pExitLocalVars->pszTargetNameDN,
	      pSecData->PartnerGSSName,
	      pChannelName) != MQXCC_OK)
    {
      gss_release_buffer (&minor_status, &output_token);
      pParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
      return;
    }
  }
#endif

  /* major_status is checked in composeSecTokenMsg */

  composeSecTokenMsg (pParms,
	     major_status,
	     minor_status,
	     GSS_SEC_ACC_SEC,
	     &output_token,
	     pOutputMsgAddr,
	     pOutputMsgLength,
	     pChannelName);

  gss_release_buffer (&minor_status, &output_token);

}



/**********************************************************************/
/*                                                                    */
/* Function Name: terminateSecurityExit                               */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*           Terminate the security context at end of the security    */
/*           exit call.                                               */
/*                                                                    */
/* Input Parameters: pContext,                                        */
/*                                                                    */
/* Output Parameters: pParms.                                         */
/*                                                                    */
/* InOut Parameters:  pSecData.                                       */
/*                                                                    */
/* Returns: None                                                      */
/*                                                                    */
/**********************************************************************/

void terminateSecurityExit (PMQCXP               pParms,
	      PSECEXITDATA pSecData,
	      PCONTEXTDATA pContext)
{
  PEXITUSERAREA   pExitUserArea   = (PEXITUSERAREA)&(pParms->ExitUserArea);
  PEXITLOCALVARS  pExitLocalVars  = pExitUserArea->pExitLocalVars;

  OM_uint32               minor_status    = 0;
  OM_uint32               major_status    = 0;
  gss_buffer_desc output_token    = {0, NULL};

  dtkDebug (pExitLocalVars, "terminateSecurityExit():\n");

  if (pSecData != NULL)
  {
    if (pSecData->PartnerGSSName != NULL)
    {
      major_status = gss_release_name (&minor_status, &(pSecData->PartnerGSSName));
    } /* endif */

    if (pSecData->credentialHandle != (gss_cred_id_t)NULL)
    {
      dtkDebug (pExitLocalVars, "terminateSecurityExit(): Calling \"gss_release_cred()\"\n");
      major_status = gss_release_cred (&minor_status, &(pSecData->credentialHandle));
    } /* endif */
  }

  if (pContext->contextHandle != (gss_ctx_id_t)NULL)
  {
    dtkDebug (pExitLocalVars, "terminateSecurityExit(): Calling \"gss_delete_sec_context()\"\n");
    major_status = gss_delete_sec_context (&minor_status,
	            &(pContext->contextHandle),
	            &output_token);

    if (major_status == GSS_S_COMPLETE)
    {
      gss_release_buffer (&minor_status, &output_token);
    } /* endif */
  }

  freeExitUserArea (pParms);

  pParms->ExitResponse = MQXCC_OK; // This was in sample....
}


/**********************************************************************/
/*                                                                    */
/* Function Name: compareNames                                        */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*           Compare the Target and Partner names                     */
/*                                                                    */
/* Input Parameters: pTargetName                                      */
/*                   pPartnerName                                     */
/*                                                                    */
/* Output Parameters: None                                            */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: MQLONG    return code value                               */
/*                                                                    */
/**********************************************************************/

MQLONG compareNames (PEXITLOCALVARS     pExitLocalVars,
	   PMQCHAR                pTargetName,
	 PMQCHAR            pPartnerName)
{
   MQLONG rc;

   dtkDebug (pExitLocalVars, "compareNames(<%s>, <%s>):\n",
      pTargetName ? pTargetName : "null",
      pPartnerName ? pPartnerName : "null");

   if ((pPartnerName != NULL) && (strlen ((char *)pPartnerName) > 0))
   {
      if (*pTargetName == '\0')
      {
   rc = MQXCC_OK;
      }
      else
      {
   if (strcmp ((char *)pTargetName, (char *)pPartnerName) == 0)
   {
     rc = MQXCC_OK;
   }
   else
   {
     rc = MQXCC_SUPPRESS_FUNCTION;
   }
      } /* endif */
   }
   else
   {
      rc = MQXCC_SUPPRESS_FUNCTION;
      /* diagnostics are output from the calling routine */

   } /* endif */

   return rc;
}



/**********************************************************************/
/*                                                                    */
/* Function Name: verifyTargetName                                    */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*           Verify the target name against the remote partner name.  */
/*           This allows restricted channels to be supported.         */
/*           It is only done in this routine for SVRCONN??channels    */
/*           because in all other cases the remote partner name is    */
/*           known when the exit is first called and is checked using */
/*           compareMQNames.                                          */
/*                                                                    */
/* Input Parameters: pParms                                           */
/*                                                                    */
/* Output Parameters: pChDef.                                         */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: MQLONG    return code value                               */
/*                                                                    */
/**********************************************************************/

MQLONG verifyTargetName (PMQCXP pParms, PMQCD pChDef)
{
  PEXITUSERAREA   pExitUserArea;
  PEXITLOCALVARS  pExitLocalVars;
  PSECEXITDATA    pSecData;

  OM_uint32       ulMajor  = 0;
  OM_uint32       ulMinor  = 0;
  gss_OID         nameType      = GSS_C_NULL_OID;
  gss_buffer_desc myFullGSSName = {0, NULL};

  pExitUserArea   = (PEXITUSERAREA)&(pParms->ExitUserArea);
  pExitLocalVars  = pExitUserArea->pExitLocalVars;
  pSecData                = pExitUserArea->pSecExitData;

  ulMajor = gss_display_name (&ulMinor, pSecData->PartnerGSSName,
	        &myFullGSSName, &nameType);

  if (IsStatusOK (pExitLocalVars, __FILE__, __LINE__, ulMajor, ulMinor,
    GSS_S_COMPLETE, GSS_S_G_NONE, "verifyTargetName()->gss_display_name()", TRUE))
  {
    PMQBYTE FullGSSName;

    dtkDebug (pExitLocalVars, "verifyTargetName(): Allocating FullGSSName "
      "for <%ld> bytes\n", myFullGSSName.length + 1);
    FullGSSName = (PMQBYTE)malloc(myFullGSSName.length + 1);

    if (FullGSSName != NULL)
    {
      MQLONG  retCode;
      PMQCHAR ClientUserid;
      int ch = '/';

      memcpy (FullGSSName, myFullGSSName.value, myFullGSSName.length);
      FullGSSName[myFullGSSName.length] = '\0';

      ClientUserid = (PMQCHAR) strrchr ((char*)FullGSSName, ch);
#ifdef GSS_DCE_API
      ClientUserid++;   /* step over / character */

      retCode = compareNames (pExitLocalVars, pSecData->mqTargetName, ClientUserid);
#else
      retCode = compareNames (pExitLocalVars, pSecData->mqTargetName, ClientUserid);
#endif

      /*************************************************************/
      /* This is a SVRCONN Channel. If there is no previously      */
      /* defined MCAUserIdentifier then copy the partner name into */
      /* the MCAUserIdentifier.                                    */
      /*************************************************************/
      if (retCode == MQXCC_OK &&
	 (pChDef->MCAUserIdentifier[0]  == ' ' ||
	  pChDef->MCAUserIdentifier[0]  == '\0'))
      {
	size_t length;

	length = strlen ((char *)ClientUserid);
	length = (length > MQ_USER_ID_LENGTH) ? MQ_USER_ID_LENGTH : length;

	memset (&(pChDef->MCAUserIdentifier[0]), 0, MQ_USER_ID_LENGTH);
	strncpy (&(pChDef->MCAUserIdentifier[0]), (char *)ClientUserid, length);
      }  /* endif */

      if (retCode != MQXCC_OK)
      {                       /* Names don't match */

      /* Diagnostic output here */

      }

      dtkDebug (pExitLocalVars, "verifyTargetName(): Freeing FullGSSName "
      "of size <%ld>\n", myFullGSSName.length + 1);
      free (FullGSSName);
      gss_release_buffer (&ulMinor, &myFullGSSName);

      return retCode;
    }
    else    /* malloc didn't work */
    {
      dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> processSecurityMessage(): "
	  "Could not allocate storage\n", __FILE__, __LINE__);
      return MQXCC_SUPPRESS_FUNCTION;
    }
  }
  else               /* gss_display_name DCE call failed */
  {
    return MQXCC_SUPPRESS_FUNCTION;
  }
}


/**********************************************************************/
/*                                                                    */
/* Function Name: processSecurityMessage                              */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*           Process a security message                               */
/*           process the message dependent upon the value in :-       */
/*                                      pInMsgAddr->header.msgType    */
/*                                                                    */
/* Input Parameters: pChDef                                           */
/*                   pInMsgAddr                                       */
/*                   lInMsgLength                                     */
/*                                                                    */
/* Output Parameters:pOutMsgAddr                                      */
/*                   pOutMsgLength                                    */
/*                                                                    */
/* InOut Parameters:  pParms.                                         */
/*                                                                    */
/* Returns: None                                                      */
/*                                                                    */
/**********************************************************************/

void processSecurityMessage (PMQCXP     pParms,
	       PMQCD      pChDef,
	       PGSSSECMSG pInMsgAddr,
	       MQLONG     lInMsgLength,
	       PMQPTR     pOutMsgAddr,
	       PMQLONG    pOutMsgLength)
{
  PMQBYTE      pMyName = NULL;
  PSECEXITDATA pSecData;
  PCONTEXTDATA pContext;

  PEXITUSERAREA  pExitUserArea;
  PEXITLOCALVARS pExitLocalVars;

  pExitUserArea   = (PEXITUSERAREA)&(pParms->ExitUserArea);
  pExitLocalVars  = pExitUserArea->pExitLocalVars;
  pSecData                = pExitUserArea->pSecExitData;
  pContext                = (PCONTEXTDATA)&(pChDef->SecurityUserData);

  dtkDebug (pExitLocalVars,
	     "processSecMsg(): sizeof secmsg %d offset %d inMsgLength %d\n",
	     sizeof(*pInMsgAddr),
	     offsetof(GSSSECMSG,msg),
	     lInMsgLength);

  if ((!SEC_MECHANISM(pInMsgAddr->header.mechanism)) ||
    (UNEXPECTEDMSG(pSecData, pInMsgAddr->header.msgType)))
  {       /* Invalid message received */
    if (IAMIDLE(pSecData))
    {
      pParms->ExitResponse = MQXCC_OK;
      return;
    }
    else
    {
      dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> processSecurityMessage(): "
	    "Invalid message received - pInMsgAddr->header.mechanism "
	    "<%02.2x> pInMsgAddr->header.msgType <%ld>\n",
	    __FILE__, __LINE__, pInMsgAddr->header.mechanism,
	    pInMsgAddr->header.msgType);
      pParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;

      return;
    }  /* endif */
  } /* endif */

  switch (pInMsgAddr->header.msgType)
  {
    case GSS_SEC_INIT_SEC:

      /* acceptor: need to check partner name and create the credentials if first time through */
      if (pSecData->credentialHandle == (gss_cred_id_t)NULL)
      {
	/* if necessary and possible check target name against partner name in exit parameters */
	pParms->ExitResponse =
	  compareMQNames (pExitLocalVars,
	          pExitLocalVars->mqTargetName,
	          &(pParms->PartnerName[0]),
	          pChDef->ChannelType,
	          (PMQCHAR)&(pChDef->ChannelName),
	          (PMQCHAR)&(pChDef->QMgrName));

	if (pParms->ExitResponse == MQXCC_OK)
	  pParms-> ExitResponse =
	    createCreds (pParms,
	           pChDef,
	           &(pSecData->credentialHandle),
	           &(pContext->contextHandle) );
      }

      if (pParms->ExitResponse == MQXCC_OK)
	acceptSecContext (pParms,
	          pSecData,
	          pContext,
	          (MQBYTE *)&(pInMsgAddr->msg),
	          lInMsgLength- offsetof(GSSSECMSG,msg) ,
	          (PMQBYTE *)pOutMsgAddr,
	          pOutMsgLength,
	          (PMQCHAR)&(pChDef->ChannelName));
      break;
    case GSS_SEC_ACC_SEC:
      initSecContext (pParms,
	        pSecData,
	        pContext,
	        (MQBYTE *)&(pInMsgAddr->msg),
	        lInMsgLength- offsetof(GSSSECMSG,msg),
	        (PMQBYTE *)pOutMsgAddr,
	        pOutMsgLength,
	        (PMQCHAR)&(pChDef->ChannelName));
      break;
  } /* endswitch */

  if ((pParms->ExitResponse == MQXCC_OK) ||
    (pParms->ExitResponse == MQXCC_SEND_SEC_MSG)) /* fully authorized */
  {
    if (pChDef->ChannelType == MQCHT_SVRCONN)
    {
      /* only do restricted channel check here if partner is an MQ client */

      if (verifyTargetName (pParms, pChDef) != MQXCC_OK)
      {
	dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> processSecurityMessage(): "
	    "-- SECURITY VIOLATION -- Could not verify TargetName for channel type <MQCHT_SVRCONN>\n"
	    "<%02.2x> pInMsgAddr->header.msgType <%ld>\n",
	    __FILE__, __LINE__);
	pParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
	return;
      }
    }  /* done verifyTargetName if needed */

    EXPECTSECMSG(pParms->ExitResponse,
	   pSecData,
	   GSS_SEC_NONE);

#if GSS_DCE_API
    pContext->mechanism = GSS_DCE_MACHANISM;  /* for snd/rcv/msg exits */
#elif GSS_ENTRUST_API
    pContext->mechanism = GSS_ENTRUST_MECHANISM;
#else
    #error No GSS_ENTRUST_API or GSS_DCE_API preprocessor defined
#endif
  } /* end of code run when this partner is fully authorized */
}

/**********************************************************************/
/*                                                                    */
/* Function Name: sealData                                            */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*           Sign the data with a digital signiture.  Optionally,     */
/*           encrypt the data.                                        */
/*                                                                    */
/* Input Parameters:  pParms                                          */
/*                                                                    */
/* Output Parameters: pChDef.                                         */
/*                                                                    */
/* InOut Parameters:  None.                                           */
/*                                                                    */
/* Returns: MQLONG    return code value                               */
/*                                                                    */
/**********************************************************************/

MQLONG sealData (PEXITLOCALVARS         pExitLocalVars,
	 PMQCXP                         pParms,
	 gss_ctx_id_t           hContext,
	 gss_buffer_desc*       pInBuffer,
	 gss_buffer_desc*       pOutBuffer)
{
  OM_uint32       ulMajor  = 0;
  OM_uint32       ulMinor  = 0;
  int                     iConfState;

  ++(pExitLocalVars->ulWrapMessagesCounter);

  if (pExitLocalVars->cDebug != EAMS_DEBUG_NONE)
  {
    int             iUnknownOffset  = 40;  // undocumented TSH structure
    int             iMsgHeaderLen   = sizeof (MQXQH) + iUnknownOffset;

    char*   pDebugMsg;
    int             iMsgLen;
    int             iDebugDisplayMsgLen;

    // Do not show MQXQH header
    if (MSGEXITCALL)
    {       // Display buffer as is - MSGEXIT do not prefix with headers
      pDebugMsg       = (char*)pInBuffer->value;
      iMsgLen         = (int) pInBuffer->length;
    }
    else
    {       // Do not show MQXQH header  - SENDEXIT do have headersR
      pDebugMsg       = (char*)pInBuffer->value + iMsgHeaderLen;
      iMsgLen         = (int) pInBuffer->length - iMsgHeaderLen;

    }

    if (iMsgLen > pExitLocalVars->iDebugMaxAppData)
    {
      iDebugDisplayMsgLen = pExitLocalVars->iDebugMaxAppData;
    }
    else if (iMsgLen < 0)
    {
      iMsgLen = 0;
      iDebugDisplayMsgLen = 0;
    }
    else
      iDebugDisplayMsgLen = iMsgLen;

    dtkDebug (pExitLocalVars, "sealData(): Requested to <%s> data. "
	"Displaying input message%s for "
	"the first <%d> of <%d> bytes = \"%.*s\"\n",
	pExitLocalVars->bEncrypt ? "Sign and Encrypt" : "Sign",
	MSGEXITCALL ? "" : " (@buffer + MQXQH offset)",
	iDebugDisplayMsgLen, iMsgLen,
	iDebugDisplayMsgLen, pDebugMsg);

    dtkDebug (pExitLocalVars, "sealData(): Calling \"gss_wrap()\"\n");

    dtkDebug (pExitLocalVars, "sealData(): Processed <%lu> messages\n",
	pExitLocalVars->ulWrapMessagesCounter);
  }

  ulMajor = gss_wrap (&ulMinor,                   /* minor_status                         */
	    hContext,                       /* context_handle                       */
	    (int) pExitLocalVars->bEncrypt, /* conf_req_flag*/
	    GSS_C_QOP_DEFAULT,      /* qop_req                                      */
	    pInBuffer,                      /* input_message_buffer         */
	    &iConfState,            /* conf_state                           */
	    pOutBuffer);            /* output_message_buffer        */

  if (!IsStatusOK (pExitLocalVars, __FILE__, __LINE__, ulMajor, ulMinor,
    GSS_S_COMPLETE, GSS_S_G_NONE, "sealData()->gss_wrap()", TRUE))
    return MQXCC_CLOSE_CHANNEL;

  return MQXCC_OK;
}

/*
 *
 */
MQLONG unSealData (PEXITLOCALVARS       pExitLocalVars,
	   PMQCXP                       pParms,
	   gss_ctx_id_t         hContext,
	   gss_buffer_desc*     pInBuffer,
	   gss_buffer_desc*     pOutBuffer)
{
  OM_uint32       ulMajor  = 0;
  OM_uint32       ulMinor  = 0;
  int                     iConfState;
  gss_qop_t       ulProtectionQuality;
  int                     iDebugDisplayLen =      ((int) pInBuffer->length > 20 ||
	           (int) pInBuffer->length < 0) ?
	          20 : (int) pInBuffer->length;

  dtkDebug (pExitLocalVars, "unSealData(): Requested to Verify Signature "
    "and optionally Decrypt data\n");
  dtkDebug (pExitLocalVars, "unSealData(): Calling \"gss_unwrap()\"\n");

  ++(pExitLocalVars->ulUnwrapMessagesCounter);

  ulMajor = gss_unwrap (&ulMinor,                         /* minor_status                         */
	      hContext,                             /* context_handle                       */
	      pInBuffer,                    /* input_message_buffer         */
	      pOutBuffer,                   /* output_message_buffer        */
	      &iConfState,                  /* conf_state                           */
	      &ulProtectionQuality);/* quality of protection        */

  if (!IsStatusOK (pExitLocalVars, __FILE__, __LINE__, ulMajor, ulMinor,
    GSS_S_COMPLETE, GSS_S_G_NONE, "unSealData()->gss_unwrap()", TRUE))
    return MQXCC_CLOSE_CHANNEL;

  if (pExitLocalVars->cDebug != EAMS_DEBUG_NONE)
  {
    int             iUnknownOffset  = 40; // STH structure ??
    int             iMsgHeaderLen   = sizeof (MQXQH) + iUnknownOffset;

    char*   pDebugMsg;
    int             iMsgLen;
    int             iDebugDisplayMsgLen;

    // Do not show MQXQH header
    if (MSGEXITCALL)
    {       // Display buffer as is - MSGEXIT do not prefix with headers
      pDebugMsg       = (char*)pOutBuffer->value;
      iMsgLen = (int) pOutBuffer->length;
    }
    else
    {       // Do not show MQXQH header  - RCVEXIT do have headersR
      pDebugMsg       = (char*)pOutBuffer->value + iMsgHeaderLen;
      iMsgLen         = (int) pOutBuffer->length - iMsgHeaderLen;
    }


    if (iMsgLen > pExitLocalVars->iDebugMaxAppData)
    {
      iDebugDisplayMsgLen = pExitLocalVars->iDebugMaxAppData;
    }
    else if (iMsgLen < 0)
    {
      iMsgLen = 0;
      iDebugDisplayMsgLen = 0;
    }
    else
      iDebugDisplayMsgLen = iMsgLen;

    dtkDebug (pExitLocalVars, "unSealData(): Message was <%s>\n"
	"Displaying output message%s for "
	"the first <%d> of <%d> bytes = \"%.*s\"\n",
	iConfState ? "Signature Verified and its data Decrypted" :
	"only Signature Verified",
	MSGEXITCALL ? "" : " (@buffer + MQXQH offset)",
	iDebugDisplayMsgLen, iMsgLen,
	iDebugDisplayMsgLen, pDebugMsg);

    dtkDebug (pExitLocalVars, "unSealData(): Processed <%lu> messages\n",
	pExitLocalVars->ulUnwrapMessagesCounter);
  }

  return MQXCC_OK;
}
